
from kivy.uix.widget import Widget
from kivy.uix.image import Image
from kivy.clock import Clock
from kivy.vector import Vector
from kivy.graphics import PushMatrix, Rotate, PopMatrix
from kivy.graphics.context_instructions import Color
from kivy.graphics.vertex_instructions import Line
from kivy.atlas import Atlas

from kivy.properties import (
	NumericProperty, ReferenceListProperty,
	AliasProperty, ObjectProperty, BooleanProperty,
	ListProperty, BoundedNumericProperty,
	StringProperty
)

from math import radians, sin, cos
import json

from kivygo.bounds_math import (
	peers, define_bounds, resize,
	update_bounds, aniresize, aniupdate_bounds,
	point_in_bounds, collide_bounds
)


class Rotabox(Widget):

	source_crop = StringProperty("")

	current_image = StringProperty("bounds")

	source_bounds = StringProperty("")

	atlas = ObjectProperty(None)
	

	image = ObjectProperty(None)
	'''This should be the image that any custom bounds are meant for.
	If not defined, widget will try to locate the topmost image in its tree.
	'''

	aspect_ratio = NumericProperty(0.0)
	'''The widget's aspect ratio. If not defined, image's ratio will be used.
	'''

	angle = NumericProperty(0)
	'''Angle of Rotation.
	'''


	def get_origin(self):
		return self.pivot

	def set_origin(self, point):
		angle = -radians(self.last_angle)
		orig = self.origin
		s = sin(angle)
		c = cos(angle)

		# normalize (translate point so origin will be 0,0)
		dx = point[0] - orig[0]
		dy = point[1] - orig[1]

		# un-rotate point
		xnew = dx * c - dy * s
		ynew = dx * s + dy * c

		# translate point back:
		pivot = xnew + orig[0], ynew + orig[1]

		# lock pos-pivot relation
		self.pivot_bond = [(pivot[0] - self.x) / float(self.width),
						   (pivot[1] - self.y) / float(self.height)]

		# Since the image (on canvas) always starts each frame in zero angle,
		# an [origin] change in any non-zero angle breaks the continuity of
		# motion/rotation, introducing an image translation (jump).
		# compensating by changing the widget's position.

		# prevent a bounds' update to [pos] change below.
		self.allow = 0

		# compensating for image translation
		self.pos = (self.x - (pivot[0] - point[0]),
					self.y - (pivot[1] - point[1]))

		# cannot wait for the triggered [update], at the end of this frame,
		# since it might concern other changes that require a bounds' update.
		# self.update()
		self.trigger_update()

	origin = AliasProperty(get_origin, set_origin)
	'''Sets the point of rotation. Default value is the widget's center.
	Works nicely with the [get_point] method below and points already defined in
	[custom_bounds].
	'''


	allow_rotabox = BooleanProperty(True)
	'''Enables widget's advanced collision detection. If False, widget will
	collide as a normal (non-Rotabox) widget.
	'''

	custom_bounds = ObjectProperty([[(0., 0.), (1., 0.), (1., 1.), (0., 1.)]])
	'''Custom bounds' definition interface. (See module's documentation).
	'''

	segment_mode = BooleanProperty(True)
	'''Collision detection method switch (see documentation above).
	'''

	open_bounds = ListProperty([])
	'''(segment_mode) If a polygon's index is in this list, the segment
	between the last and first points of the polygon is not considered in the
	collision checks.
	'''

	pre_check = BooleanProperty(False)
	'''(Cython) A collision optimization switch for larger widgets (45+ points).
	It's always True in Python but in Cython, for small widgets, the slight tax 
	in updating the bounds outweighs the benefit in collision.
	'''


	def get_allow_drag(self):
		return self.allow_drag_x, self.allow_drag_y

	def set_allow_drag(self, value):
		if type(value) in (list, tuple):
			self.allow_drag_x, self.allow_drag_y = value
		else:
			self.allow_drag_x = self.allow_drag_y = bool(value)

	allow_drag = AliasProperty(get_allow_drag, set_allow_drag, bind=('allow_drag_x', 'allow_drag_y'))
	'''Allow touch translation on the X or Y axis.
	'''

	allow_drag_x = BooleanProperty(False)
	'''Allow touch translation on the X axis.
	'''

	allow_drag_y = BooleanProperty(False)
	'''Allow touch translation on the Y axis.
	'''


	single_touch_rotation = BooleanProperty(False)
	'''Allow rotation around [origin].
	'''

	single_touch_scaling = BooleanProperty(False)
	'''Allow scaling around [origin].
	'''

	multi_touch_rotation = BooleanProperty(False)
	'''Allow multitouch rotation. [origin] is defined each time by the touch.
	'''

	multi_touch_scaling = BooleanProperty(False)
	'''Allow multitouch scaling. [origin] is defined each time by the touch.
	'''

	single_drag_touch = BoundedNumericProperty(1, min=1)
	'''How many touches will be treated as one single drag touch.
	'''

	single_trans_touch = BoundedNumericProperty(1, min=1)
	'''How many touches will be treated as one single rotation/scaling touch.
	'''

	touched_to_front = BooleanProperty(False)
	'''If touched, widget will be pushed to the top of parent widget tree.
	'''

	collide_after_children = BooleanProperty(False)
	'''If True, limiting the touch inside the bounds will be done after
	dispaching the touch to the child and grandchildren, so even outside the
	bounds they can still be touched.
	IMPORTANT NOTE: Grandchildren, inside or outside the bounds, can collide
		independently ONLY if widget is NOT ROTATED ([angle] must be 0).
	'''


	def get_scale(self):
		return float(self.width) / self.original_size[0]

	def set_scale(self, amount):
		if amount < self.scale_min:
			amount = self.scale_min
		elif amount > self.scale_max:
			amount = self.scale_max

		pivot = self.pivot[:]
		self.size = (amount * self.original_size[0],
					 amount * self.original_size[1])
		self.pivot = pivot
		if self.initial_scale:
			return
		self.initial_scale = self.scale
	
	scale = AliasProperty(get_scale, set_scale, bind=('width', 'height', 'original_size'))
	'''
	Widget's current scale. Calculated from [original_size] (user's input
	[size] or [image]'s [texture_size]). Used for touch scaling but it can be an
	alternative to [size].
	'''

	scale_min = NumericProperty(0.01)
	'''Minimum scale allowed.
	'''

	scale_max = NumericProperty(1e20)
	'''Maximum scale allowed.
	'''


	def get_pivot_x(self):
		return self.x + self.width * self.pivot_bond[0]

	def set_pivot_x(self, value):
		if self.width > 1:
			self.x = value - self.width * self.pivot_bond[0]
		elif value > 1:
			self.temp_piv_x = value

	def get_pivot_y(self):
		return self.y + self.height * self.pivot_bond[1]

	def set_pivot_y(self, value):
		if self.height > 1:
			self.y = value - self.height * self.pivot_bond[1]
		elif value > 1:
			self.temp_piv_y = value

	pivot_x = AliasProperty(get_pivot_x, set_pivot_x, bind=('x', 'width', 'pivot_bond'))
	pivot_y = AliasProperty(get_pivot_y, set_pivot_y, bind=('y', 'height', 'pivot_bond'))
	pivot = ReferenceListProperty(pivot_x, pivot_y)
	'''
	Point of rotation and scaling.
	While [origin] property sets [pivot]'s relation to widget, [pivot]
	itself can be used to position the widget, much like [pos] or [center].
	'''

	prepared = BooleanProperty(False)
	'''
	Its state change signifies a reset. The reset completion signal, however,
	is the consequent [ready] state change to True.
	'''

	draw_bounds = BooleanProperty(False)
	'''Enables bounds visualization (for testing).
	'''


	__events__ = ('on_transform_with_touch', 'on_touched_to_front')

	def on_transform_with_touch(self, touch):
		'''
		Called when a touch event has transformed the widget.
		By default this does nothing, but can be overriden by derived
		classes that need to react to transformations caused by user
		input.
		'''
		pass

	def on_touched_to_front(self, touch):
		'''
		Called when a touch event causes the widget to be brought to the
		front of the parent (only if :attr:`touched_to_front` is True)
		'''
		pass


	def __init__(self, **kwargs):
		self.size = [1, 1]
		self.temp_piv_x = self.temp_piv_y = 0
		self.initial_scale = 0
		self.touches = []
		self.last_touch_pos = {}
		
		super().__init__(**kwargs)
		
		self.image = Image()
		self.sized_by_img = False
		self.ratio = 1
		self.last_pos = [0, 0]
		self.last_size = self.size[:]
		self.last_angle = 0
		self.anim = False
		self.rid = 0
		self.curr_key = 'bounds'
		self.draw_color = Color(rgba=[0.29, 0.518, 1, 1])
		self.box_color = Color(rgba=[0.35, 0.15, 0, 1])
		self.draw_lines = []
		self.box_lines = []
		self.rotation = Rotate(angle=0, origin=self.center)
		self.canvas.before.add(PushMatrix())
		self.canvas.before.add(self.rotation)
		self.canvas.after.add(PopMatrix())

		self.trigger_update = Clock.create_trigger(self.update, -1)
		self.trigger_draw = Clock.create_trigger(self.draw, -1)

		self.bind(
			children=self.trigger_update,
			parent=self.trigger_update,
			pos=self.trigger_update,
			pos_hint=self.trigger_update,
			angle=self.trigger_update,
			image=self.on_reset,
			aspect_ratio=self.on_reset,
			custom_bounds=self.on_reset,
			allow_rotabox=self.on_reset,
			draw_bounds=self.on_reset
		)
	
	def add_widget(self, widget, **kwargs):
		if self.children:
			raise Exception('Rotabox can only have one child.')
		super().add_widget(widget)

	def on_open_bounds(self, *args):
		if self.open_bounds and not self.segment_mode:
			raise Exception('Open bounds are only applicable in Segment mode.')

	def on_source_bounds(self, *args):
		self.custom_bounds = self.read_bounds(self.source_bounds)

	def on_current_image(self, *args):
		if self.atlas == None:
			return None
		
		self.image.texture = self.atlas.textures[self.current_image]
		self.prepared = False
		self.trigger_update()

	def on_source_crop(self, *args):
		if self.source_crop == "":
			self.atlas = None
			self.image = None
			self.clear_widgets()
			return None
		
		if self.image != None and not self.image in self.children:
			self.add_widget(self.image)
		
		self.atlas = Atlas(self.source_crop)
		self.image.texture = tuple(self.atlas.textures.values())[0]
		
	def on_size(self, *args):
		self.trigger_update()

	def on_reset(self, *args):
		self.trigger_update()

	def reset_draw_bounds(self, *args):
		for line in self.draw_lines:
			self.canvas.after.remove(line)
		self.draw_lines = []

		for line in self.box_lines:
			self.canvas.after.remove(line)
		self.box_lines = []

	def prepare(self, n=None, *args):
		self.prepared = True
		tempscale = self.initial_scale

		if self.source_crop or self.image:
			try:
				self.image.texture.mag_filter = 'nearest'
			except AttributeError:
				pass

			self.image.allow_stretch = True

			# Calculating widget's size from available inputs.
			if (not (self.width - tempscale > 1 or self.height - tempscale > 1) or self.sized_by_img):
				try:
					self.original_size = self.image.texture.size
				except AttributeError:
					# If animation, texture not ready (?)
					self.original_size = self.image.size

				if not tempscale:
					tempscale = 1
				self.size = list(map(lambda n: (n * tempscale), self.original_size))

				if self.sized_by_img:
					dw = self.width - self.last_size[0]
					dh = self.height - self.last_size[1]
					self.pivot = ( (self.pivot_x - dw * 0.5), (self.pivot_y - dh * 0.5) )
			else:
				self.original_size = self.size[:]
				if tempscale:
					self.size = list(map(lambda n: (n * tempscale), self.size))
		else:
			self.original_size = self.size[:]
			if not tempscale:
				tempscale = 1
			self.size = list(map(lambda n: (n * tempscale), self.size))

		# Setting the widget's ratio.
		if not self.aspect_ratio:
			if self.source_crop and self.sized_by_img:
				self.ratio = self.image.image_ratio
			else:
				self.ratio = (self.width / float(self.height))
		else:
			self.ratio = self.aspect_ratio

		# If [size] is not specified explicitly by the user, any input values
		# for [pivot] are withheld in [temp_piv_x] and [temp_piv_y], until
		# [size] gets its values from [image]. Then, [pivot] gets the withheld
		# values, here.
		if self.temp_piv_x or self.temp_piv_y:
			self.pivot = self.temp_piv_x, self.temp_piv_y
			self.temp_piv_x = self.temp_piv_y = 0

		if self.allow_rotabox:
			# Generating a unique key for each instance
			try:
				self.rid = sorted(peers.keys())[-1] + 1
			except IndexError:
				pass

			if isinstance(self.custom_bounds, dict):
				self.curr_key = self.current_image
				self.anim = True
			
			define_bounds(self.custom_bounds, self.open_bounds, self.segment_mode, self.rid, self.pre_check)
			
			if self.draw_bounds:
				self.set_draw()
				self.bind(
					children=self.trigger_draw,
					parent=self.trigger_draw,
					pos=self.trigger_draw,
					pos_hint=self.trigger_draw,
					angle=self.trigger_draw,
				)

		self.trigger_update()

	def get_point(self, pol_index, point_index):
		'''Access a point's current position, based on [custom_bounds] structure.'''
		bounds = peers[self.rid][self.curr_key]
		index = (sum(bounds['pol_lens'][:pol_index]) + point_index) * 2
		return list(bounds['points'][index:index + 2])

	def read_bounds(self, filename, delayed=False, *args):
		'''
		Define [custom_bounds] using a rotaboxer's project file.
		To work, [size] should be already defined.
		'''
		project = None
	
		try:
			with open(filename, 'r', encoding="UTF8") as proj:
				project = json.load(proj)
		except (IOError, KeyError) as er:
			print('On read_bounds: ', er)
			return None
		
		bounds = {}
		opens = []
		for f, frame in project.items():
			if f in ('image', 'version'):
				continue

			pols = []
			i = 0
			while i < len(frame):
				for pol in frame.values():
					if pol['number'] == i:
						pols.append(pol)
						break
				i += 1
			bounds[f] = []
			for p, pol in enumerate(pols):
				try:
					if pol['open']:
						opens.append(p)
				except KeyError:
					pass
				
				bounds[f].append([(round(float(point[0]) / self.width, 3),
									round(float(point[1]) / self.height, 3))
									for point in pol['points']])

		blen = len(bounds)
		if blen:
			if opens:
				self.open_bounds = opens

			if blen == 1:
				bounds = bounds[list(bounds.keys())[0]]

			if delayed:
				self.custom_bounds = bounds
				return None

			return bounds

		return self.custom_bounds


	def update_size(self, *args):
		
		width, height = self.size
		if round(self.ratio, 3) != round(width / float(height), 3):
			# Adjusting size to fit ratio
			last_size = self.last_size
			if abs(width - last_size[0]) > abs(height - last_size[1]):
				self.size = [width, (float(width) / self.ratio)]
			else:
				self.size = [(height * self.ratio), height]
			
			if last_size != [1, 1]:
				dw = self.width - last_size[0]
				dh = self.height - last_size[1]
				# Moving widget to keep pivot's position the same, to originate the resizing from pivot.
				self.pivot = ( (self.pivot_x - dw * 0.5), (self.pivot_y - dh * 0.5) )

		if self.allow_rotabox:
			# Scaling widget's bounds
			if not self.anim:
				resize(self.width, self.height, self.rid)
				update_bounds(self.pos, radians(self.angle), self.origin, self.rid, self.curr_key)
				self.allow = 0
			else:
				aniresize(self.width, self.height, self.rid)

		if self.children:
			self.children[0].size = self.size

		self.last_size = self.size[:]
		self.trigger_update()

	def update(self, *args):
		'''Updates the widget's angle, point of rotation, bounds and child's
		position.
		Also runs the [update_size] method on size change and the [prepare]
		method initially and on reset.
		'''
		if not self.prepared:
			self.prepare()
			self.update_size()
			return None

		self.angle %= 360
		angle = self.angle
		pos = self.pos

		# Updating the rotation instruction, in canvas.before
		self.rotation.origin = self.origin
		self.rotation.angle = angle

		# Updating the child's position
		if self.children:
			self.children[0].pos = pos

		motion = [pos[0] - self.last_pos[0], pos[1] - self.last_pos[1]]
		if abs(motion[0]) < .01 and abs(motion[1]) < 0.01:
			motion = []
		self.last_pos = pos[:]

		angle_diff = angle - self.last_angle
		if abs(angle_diff) < 0.01:
			angle_diff = 0

		self.last_angle = angle
		if not self.allow:
			self.allow = 1
			return None

		if self.allow_rotabox:
			# Updating the custom bounds
			if self.anim:
				# An identically keyed atlas file is assumed.
				self.curr_key = self.current_image
				aniupdate_bounds(True, pos, radians(angle), self.origin, self.rid, self.curr_key)
				return None

			if motion or angle_diff:
				update_bounds(motion, radians(angle_diff), self.origin, self.rid, self.curr_key)
	

	def collide_point(self, x=0, y=0):
		if self.allow_rotabox:
			return point_in_bounds(x, y, self.rid, frame=self.curr_key)
		else:
			return super().collide_point(x, y)

	def collide_widget(self, wid):
		try:
			try:
				return collide_bounds(self.rid, wid.rid, frame=self.curr_key, tframe=wid.curr_key)
			except AttributeError:
				if self.segment_mode:
					return collide_bounds(self.rid, [wid.x, wid.y, wid.right, wid.top], frame=self.curr_key)
				return super().collide_widget(wid)
		except KeyError:
			return False

	def on_touch_down(self, touch):
		x, y = touch.x, touch.y

		# if the touch isnt on the widget we do nothing
		if not self.collide_after_children:
			if not self.collide_point(x, y):
				return False

		# let the child widgets handle the event if they want
		if super().on_touch_down(touch):
			# ensure children don't have to do it themselves
			if 'multitouch_sim' in touch.profile:
				touch.multitouch_sim = True
			
			self.bring_to_front(touch)
			return True

		# if we don't have any active
		# interaction control, then don't accept the touch.
		if not (self.allow_drag_x
				or self.allow_drag_y
				or self.multi_touch_rotation
				or self.multi_touch_scaling
				or self.single_touch_rotation
				or self.single_touch_scaling):
			return False

		if self.collide_after_children:
			if not self.collide_point(x, y):
				return False

		if 'multitouch_sim' in touch.profile:
			touch.multitouch_sim = True

		# grab the touch so we get all it later move events for sure
		self.bring_to_front(touch)
		touch.grab(self)
		self.touches.append(touch)
		self.last_touch_pos[touch] = x, y
		return True

	def bring_to_front(self, touch):
		if self.touched_to_front and self.parent:
			parent = self.parent
			if parent.children[0] is self:
				return None
			
			parent.remove_widget(self)
			parent.add_widget(self)
			self.dispatch('on_touched_to_front', touch)

	def on_touch_move(self, touch):
		x, y = touch.x, touch.y

		# let the child widgets handle the event if they want
		if self.collide_point(x, y) and not touch.grab_current == self:
			if super().on_touch_move(touch):
				return True

		if touch in self.touches and touch.grab_current == self:
			if self.transform_with_touch(touch):
				self.dispatch('on_transform_with_touch', touch)
			self.last_touch_pos[touch] = x, y

		if self.collide_point(x, y):
			return True

	def transform_with_touch(self, touch):
		changed = False
		touches = len(self.touches)

		if (touches == self.single_drag_touch and (self.allow_drag_x or self.allow_drag_y)):
			dx = (touch.x - self.last_touch_pos[touch][0]) * self.allow_drag_x
			dy = (touch.y - self.last_touch_pos[touch][1]) * self.allow_drag_y
			
			self.x += float(dx) / self.single_drag_touch
			self.y += float(dy) / self.single_drag_touch
			changed = True

		if (touches == self.single_trans_touch and (self.single_touch_rotation or self.single_touch_scaling)):
			anchor = self.pivot
			old_line = Vector(*touch.ppos) - anchor
			new_line = Vector(*touch.pos) - anchor
			if not old_line.length():
				return changed
			
			if self.single_touch_rotation:
				angle = new_line.angle(old_line)
				self.angle += angle

			if self.single_touch_scaling:
				_scale = new_line.length() / old_line.length()
				self.scale *= _scale
			
			changed = True

		if touches == 1:
			return changed

		# we have more than one touch... list of last known pos
		points = [Vector(self.last_touch_pos[t]) for t in self.touches if t is not touch]
		points.append(Vector(touch.pos))

		# we only want to transform if the touch is part of the two touches
		# farthest apart! So first we find anchor, the point to transform
		# around as another touch farthest away from current touch's pos
		anchor = max(points[:-1], key=lambda p: p.distance(touch.pos))

		self.origin = anchor

		# now we find the touch farthest away from anchor, if its not the
		# same as touch. Touch is not one of the two touches used to transform
		farthest = max(points, key=anchor.distance)
		if farthest is not points[-1]:
			return changed

		# ok, so we have touch, and anchor, so we can actually compute the transformation
		old_line = Vector(*touch.ppos) - anchor
		new_line = Vector(*touch.pos) - anchor
		if not old_line.length():
			return changed

		angle = new_line.angle(old_line) * self.multi_touch_rotation
		self.angle += angle

		if self.multi_touch_scaling:
			_scale = new_line.length() / old_line.length()
			self.scale *= _scale
			changed = True
		
		return changed

	def on_touch_up(self, touch):
		# if the touch isnt on the widget we do nothing, just try children
		if not touch.grab_current == self:
			if super().on_touch_up(touch):
				return True

		# remove it from our saved touches
		if touch in self.touches and touch.grab_state:
			touch.ungrab(self)
			del self.last_touch_pos[touch]
			self.touches.remove(touch)

		# stop propagating if its within our bounds
		if self.collide_point(*touch.pos):
			return True

	def set_draw(self, *args):
		'''Setting up canvas for test-drawing the bounds.'''
		self.reset_draw_bounds()
		
		if self.anim:
			pols = max([len(frame) for frame in self.custom_bounds.values()])
			
			length = max([len(pol) for pol in self.custom_bounds.values()])
		else:
			pols = len(self.custom_bounds)
			length = peers[self.rid][self.curr_key]['length']

		for i in range(pols):
			if i not in self.open_bounds or not self.segment_mode:
				self.draw_lines.append(Line(close=True, dash_offset=3, dash_length=5))
			else:
				self.draw_lines.append(Line(close=False, dash_offset=3, dash_length=5))

		self.canvas.after.add(self.draw_color)
		for line in self.draw_lines:
			self.canvas.after.add(line)

		# Securing draw on a stationary widget.
		self.x += 0.001
		self.trigger_draw()

	def draw(self, *args):
		''' If [draw_bounds] is True, visualises the widget's bounds. For testing.
		''' 
		if not self.prepared:
			return None
		
		try:
			bounds = peers[self.rid][self.curr_key]
		except KeyError:
			return None

		start = 0
		for i, leng in enumerate(bounds['pol_lens']):
			self.draw_lines[i].points = [bounds['points'][j] for j in range(start, start + leng * 2)]
			start += leng * 2


	def get_size_hint_x(self):
		# Locking size_hint property to None, None,
		# in order to keep intended aspect ratio (critical for custom bounds).
		return None

	def set_size_hint_x(self, value):
		raise Exception("Rotabox can't use size_hint.")
	
	size_hint_x = AliasProperty(get_size_hint_x, set_size_hint_x)


	def get_size_hint_y(self):
		return None

	def set_size_hint_y(self, value):
		raise Exception("Rotabox can't use size_hint.")

	size_hint_y = AliasProperty(get_size_hint_y, set_size_hint_y)
	size_hint = ReferenceListProperty(size_hint_x, size_hint_y)


	original_size = ListProperty([1, 1])
	"""
	Used for calculating the widget's scale. It's the user's input size
	or the [texture_size] of the widget's [image].
	"""

	pivot_bond = ListProperty([.5, .5])
	"""
	A fractional value that keeps [pivot] relative to the widget's size
	and position. [origin] sets and changes it.
	"""

	allow = BooleanProperty(1)
	""" Switch to suspend bounds' update while repositioning, during an [origin] change or a resize.
	"""



import os, json
from kivy.properties import (
	StringProperty, ColorProperty,
	ObjectProperty
)
from kivy.utils import get_color_from_hex
from kivy.app import App
from kivygo.utils import do_correction_path
from kivygo import colors


class kivygoApp(App):

	path_json = StringProperty("pallet")
	theme_json = ObjectProperty(None)

	root_path = StringProperty("")

	icon_path = StringProperty("assets/icons")

	image_path = StringProperty("assets/images")

	font_path = StringProperty("fonts")

	current_pallet = ObjectProperty(None)
	
	_app_file = StringProperty(os.path.split(__file__)[0])

	def __init__(self, *args, **kwargs):
		# Set all colors properties to the App accordian to the `colors.PALLET_KEY_COLORS`
		if self.current_pallet == None:
			self.current_pallet = colors.Light
		
		for _key in colors.PALLET_KEY_COLORS:
			color = getattr(self.current_pallet, _key)
			self.apply_property(
				**{ _key : ColorProperty(color) }
			)
		
		super().__init__(*args, **kwargs)
	
	def on_current_pallet(self, *args):
		self.change_pallet(self.current_pallet)

	def get_json(self, name, path=None, *args):
		if path == None:
			path = self.path_json

		with open(self.get_path(f"{path}/{name}.json"), 'r', encoding='utf-8') as file:
			return json.load(file)


	def update_json(self, new_json, name, path=None):
		if path == None:
			path = self.path_json
		
		with open(self.get_path(f"{path}/{name}.json"), 'w', encoding='utf-8') as file:
			file.write(json.dumps(new_json, indent=4))

	def on_theme_json(self, *args):
		if isinstance(self.theme_json, str):
			json_dict = self.get_json(self.theme_json)
		
		elif isinstance(self.theme_json, dict):
			json_dict = self.theme_json
		
		path = do_correction_path(self._app_file)
		default_json = self.get_json("default_json_theme", path)
		
		for key, value in json_dict:
			if key not in default_json:
				raise TypeError("Json theme has an inexistent key!")
			
			if isinstance(value, str):
				setattr(self, key, get_color_from_hex(value))
			else:
				setattr(self, key, value)

	def on_root_path(self, *args):
		self.root_path = do_correction_path(self.root_path)


	def get_icon(self, name, ext='png'):
		return self.get_path(f'{self.icon_path}/{name}.{ext}')
	

	def get_image(self, name, ext='png'):
		return self.get_path(f'{self.image_path}/{name}.{ext}')


	def get_font(self, path, name, ext="ttf"):
		return self.get_path(f'{self.font_path}/{path}/{name}.{ext}')


	def get_path(self, local, root_path=None):
		if root_path == None:
			self.root_path
			
		return f'{root_path}/{do_correction_path(local)}'

	def change_pallet(self, _obj):
		if isinstance(_obj, dict):
			return None
		
		for key in colors.PALLET_KEY_COLORS:
			if hasattr(_obj, key):
				setattr(self, key, getattr(_obj, key))
			
		self.current_pallet = _obj



"""
To use this class, you must define two methods for it:
:attr:`HoverBehavior.on_cursor_enter` and :attr:`HoverBehavior.on_cursor_leave`,
which will be automatically called when the mouse cursor is over the widget
and when the mouse cursor goes beyond the widget.

.. note::

	:class:`~HoverBehavior` will by default check to see if the current Widget is visible
	(i.e. not covered by a modal or popup and not a part of a Relative Layout, MDTab or Carousel
	that is not currently visible etc) and will only issue events if the widget is visible.

	To get the legacy behavior that the events are always triggered, you can set `detect_visible
	on the Widget to `False`.

"""
from kivy.uix.widget import Widget
from kivy.core.window import Window
from kivy.properties import (
	BooleanProperty, ObjectProperty,
	ListProperty
)


class HoverBehavior(Widget):

	hovering = BooleanProperty(False)
	"""
	`True`, if the mouse cursor is within the borders of the widget.

	Note that this is set and cleared even if the widget is not visible
	"""

	hover_visible = BooleanProperty(False)
	"""
	`True` if hovering is True AND is the current widget is visible
	"""

	enter_point = ObjectProperty(allownone=True)
	"""
	Holds the last position where the mouse pointer crossed into the Widget
	if the Widget is visible and is currently in a hovering state
	"""

	detect_visible = BooleanProperty(True)
	"""
	Should this widget perform the visibility check?
	"""

	cursor_pos = ListProperty([0, 0])
	repeat_callback = BooleanProperty(False)

	_last_parent = ObjectProperty(None)
	__resizes = []


	def on_cursor_enter(self, *args):
		"""Called when mouse enters the bbox of the widget AND the widget is visible."""
		pass

	def on_cursor_leave(self, *args):
		"""Called when the mouse exits the widget AND the widget is visible."""
		pass


	def __init__(self, **kwargs):
		super().__init__(**kwargs)

		self.register_event_type("on_cursor_enter")
		self.register_event_type("on_cursor_leave")

		if not HoverBehavior.__resizes:
			Window.bind(mouse_pos=self.on_mouse_update)
			Window.bind(on_cursor_leave=self.window_cursor_leave)
		
		HoverBehavior.__resizes.append(self)

	def on_parent(self, *args):
		if self.parent != None:
			self._last_parent = self.parent
		
		if self.parent == None and HoverBehavior.__resizes:
			if HoverBehavior.__resizes[0] is self:
				Window.unbind(mouse_pos=self.on_mouse_update)
				Window.unbind(on_cursor_leave=self.window_cursor_leave)
				del HoverBehavior.__resizes[0]
				
				if HoverBehavior.__resizes:
					Window.bind(mouse_pos=HoverBehavior.__resizes[0].on_mouse_pos)
					Window.bind(on_cursor_leave=HoverBehavior.__resizes[0].window_cursor_leave)
			else:
				HoverBehavior.__resizes.remove(self)

	def window_cursor_leave(self, *args):
		for wid in HoverBehavior.__resizes:
			if wid.hover_visible or wid.repeat_callback:
				wid.do_cursor_leave()

	def do_cursor_leave(self, *args):
		self.hovering = False
		self.enter_point = None
		if self.hover_visible or self.repeat_callback:
			self.hover_visible = False
			self.dispatch("on_cursor_leave")
	
	def do_cursor_enter(self, *args):
		if self.hover_visible:
			self.enter_point = self.cursor_pos
			self.dispatch("on_cursor_enter")

	def new_dispatch(self, *args):
		"""
			ARGS:
			 `option`: 'leave' or 'enter'
		"""
		if not self.hovering:
			self.dispatch("on_cursor_leave")
		else:
			self.dispatch("on_cursor_enter")

	def on_mouse_update(self, *args):
		if not self.get_root_window():
			return None
		
		for wid in HoverBehavior.__resizes:
			
			pos = args[1]
			wid.cursor_pos = pos
			#  If widget not in the same position: on_exit event if needed
			if not wid.collide_point(*wid.to_widget(*pos)):
				wid.do_cursor_leave()
				continue

			# | The pointer is in the same position as the widget

			if wid.hovering and not wid.repeat_callback:
				#  nothing to do here. Not - this does not handle the case where
				#  a popup comes over an existing hover event.
				#  This seems reasonable
				continue
			
			wid.hovering = True

			# | We need to traverse the tree to see if the Widget is visible
			
			# This is a two stage process:
			# - first go up the tree to the root Window.
			#   At each stage - check that the Widget is actually visible
			# - Second - At the root Window check that there is not another branch
			#   covering the Widget

			wid.hover_visible = True
			if wid.detect_visible:
				widget = wid
				while 1:
					# Walk up the Widget tree from the target Widget
					parent = widget.parent
					try:
						# See if the mouse point collides with the parent
						# using both local and global coordinates to cover absolut and relative layouts
						pinside = parent.collide_point(*parent.to_widget(*pos)) or parent.collide_point(*pos)
					except Exception:
						# The collide_point will error when you reach the root Window
						break
					if not pinside:
						wid.hover_visible = False
						break
					# Iterate upwards
					widget = parent

				#  parent = root window
				#  widget = first Widget on the current branch
				children = parent.children if parent != None else wid.children
				for child in children:
					# For each top level widget - check if is current branch
					# If it is - then break.
					# If not then - since we start at 0 - this widget is visible

					# Check to see if it should take the hover
					if child == widget:
						# this means that the current widget is visible
						break
					if child.collide_point(*pos):
						#  this means that the current widget is covered by a modal or popup
						wid.hover_visible = False
						break
					
			wid.do_cursor_enter()





from kivy.lang import global_idmap
from kivy.properties import (
    ObjectProperty,
    OptionProperty,
    StringProperty,
)
from kivy.uix.effectwidget import AdvancedEffectBase, EffectWidget
from kivy.graphics import BindTexture

GLSL = """

uniform sampler2D mask;
uniform vec2 mask_pos;
uniform vec2 mask_size;

// 0: the mask pixel multiply the texture
// 1: the mask pixel substract to the texture
// addition and division don't seem to be of any use
uniform int mode;
vec4 effect(vec4 color, sampler2D texture, vec2 tex_coords, vec2 coords){
    // get the pixel lookup pos in the mask relative to our main texture
    vec2 pos = tex_coords; // - mask_pos * resolution;
    vec2 size = mask_size / resolution;
    if (
        pos.x < 0. || pos.y < 0. ||
        pos.x > 1. || pos.y > 1.
    )
        if (mode == 0)
            return vec4(0.);
        else
            return color;
    if (mode == 0)
        return color * texture2D(mask, pos);
    else
        return color - texture2D(mask, pos);
}
"""


class MaskEffect(AdvancedEffectBase):
    mask = ObjectProperty(None)
    glsl = StringProperty(GLSL)
    mode = OptionProperty("multiply", options=["multiply", "substract"])

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        assert self.mask
        self.mask.bind(
            pos=self.update_mask,
            size=self.update_mask,
        )
        self.bind(mode=self.update_mask)

    def on_fbo(self, *args):
        self.update_mask()

    def update_mask(self, *args):
        with self.fbo:
            BindTexture(texture=self.mask.fbo.texture, index=1)
        
        self.uniforms.update(
            {
                "mask": 1,
                "mask_pos": list(self.mask.pos),
                "mask_size": list(self.mask.size),
                "mode": 0 if self.mode == "multiply" else 1,
            }
        )


global_idmap["MaskEffect"] = MaskEffect




from kivy.lang import Builder
from kivy.uix.widget import Widget
from kivy.properties import ListProperty, NumericProperty
from kivy.metrics import dp
import numpy as np


Builder.load_string('''

#:import chain itertools.chain


<BezierLine>:
	_points: list(chain(*self.points))

	canvas:
		Color:
			rgba: [1, 1, 1, 0.2]
		SmoothLine:
			points: self._points

		Color:
			rgba: [1, 1, 1, 1]
		Line:
			bezier: self._points
			width: 2

		Color:
			rgba: [1, 1, 1, 0.5]
		Point:
			points: self._points
			pointsize: 5

''')


def dist(a, b):
	return ( ((a[0] - b[0]) ** 2 + (a[1] - b[1]) ** 2) ** 0.5 )


class BezierLine(Widget):

	_points = ListProperty([])
	points = ListProperty([])
	select_dist = NumericProperty(10)
	delete_dist = NumericProperty(5)

	def on_touch_down(self, touch):
		if super().on_touch_down(touch):
			return True

		max_dist = dp(self.select_dist)

		for i, p in enumerate(self.points):
			if dist(touch.pos, p) < max_dist:
				touch.ud['selected'] = i
				touch.grab(self)
				return True

		for i, p in enumerate(self.points[:-1]):
			index = (i + 1)
			new_dist = dist(touch.pos, p) + dist(touch.pos, self.points[index])
			new_dist -= dist(p, self.points[index])
			
			if (new_dist < max_dist):
				self.points = self.points[:index] + [list(touch.pos)] + self.points[index:]
				touch.ud['selected'] = index
				touch.grab(self)
				return True

	def on_touch_move(self, touch):
		if touch.grab_current is not self:
			return super().on_touch_move(touch)
		
		point = touch.ud['selected']
		self.points[point] = touch.pos

	def on_touch_up(self, touch):
		if touch.grab_current is not self:
			return super().on_touch_up(touch)
		
		touch.ungrab(self)
		i = touch.ud['selected']
		
		if touch.is_double_tap:
			if len(self.points) < 3:
				self.parent.remove_widget(self)
			elif i <= 0:
				self.points = self.points[i + 1::]
			else:
				self.points = self.points[::i] + self.points[i + 1::]


class Bezier():

	def TwoPoints(t, P1, P2):
		"""
		Returns a point between P1 and P2, parametised by t.
		INPUTS:
			t     float/int; a parameterisation.
			P1    numpy array; a point.
			P2    numpy array; a point.
		OUTPUTS:
			    numpy array; a point.
		"""

		if not isinstance(P1, np.ndarray) or not isinstance(P2, np.ndarray):
			raise TypeError("Points must be an instance of the numpy.ndarray!")
		
		if not isinstance(t, (int, float)):
			raise TypeError("Parameter t must be an int or float!")

		return ( (1 - t) * P1 + t * P2 )

	def Points(t, points):
		"""
		Returns a list of points interpolated by the Bezier process
		INPUTS:
			t            float/int; a parameterisation.
			points       list of numpy arrays; points.
		OUTPUTS:
			newpoints    list of numpy arrays; points.
		"""

		newpoints = []
		for i1 in range(len(points) - 1):
			newpoints += [Bezier.TwoPoints(t, points[i1], points[i1 + 1])]

		return newpoints

	def Point(t, points):
		"""
		Returns a point interpolated by the Bezier process
		INPUTS:
			t            float/int; a parameterisation.
			points       list of numpy arrays; points.
		OUTPUTS:
			newpoint     numpy array; a point.
		"""

		newpoints = points
		while len(newpoints) > 1:
			newpoints = Bezier.Points(t, newpoints)

		return newpoints[0]

	def Curve(t_values, points):
		"""
		Returns a point interpolated by the Bezier process
		INPUTS:
			t_values     list of floats/ints; a parameterisation.
			points       list of numpy arrays; points.
		OUTPUTS:
			curve        list of numpy arrays; points.
		"""

		if not hasattr(t_values, "__iter__") \
			or not isinstance(t_values[0], (int, float)) \
			or len(t_values) < 1:
			
			raise TypeError("`t_values` Must be an iterable of integers or floats, of length greater than 0 .")
		
		
		curve = np.array([[0.0] * len(points[0])])
		for t in t_values:
			curve = np.append(curve, [Bezier.Point(t, points)], axis=0)

		curve = np.delete(curve, 0, 0)
		return curve




from kivy.uix.boxlayout import BoxLayout
from kivygo.behaviors.hover import HoverBehavior
from kivygo.behaviors.button import ButtonBehavior
from kivygo.behaviors.drag_and_drop import DraggableLayoutBehavior
from kivy.lang.builder import Builder
from kivy.clock import Clock
from kivy.properties import ListProperty, NumericProperty
from kivy.metrics import dp


Builder.load_string("""

<ColoredBoxLayout>:
	canvas.before: 
		Color:
			rgba: self.background_color
		RoundedRectangle:
			pos: self.pos
			size: self.size
			radius: self.radius
		Color:
			rgba: [1, 1, 1, 1]
	canvas.after:
		Color:
			rgba: self.stroke_color
		Line:
			rounded_rectangle: [*self.pos, *self.size, *self.radius]
			width: self.stroke_width
		Color:
			rgba: [1, 1, 1, 1]

""")


class ColoredBoxLayout(BoxLayout, HoverBehavior):

	background_color = ListProperty([0, 0, 0, 0])
	radius = ListProperty([0, 0, 0, 0])
	stroke_color = ListProperty([0, 0 ,0 ,0])
	stroke_width = NumericProperty(dp(2))


class ColoredButtonBoxLayout(ButtonBehavior, ColoredBoxLayout):
	
	background_state_color = ListProperty([-1, -1, -1, -1])

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		Clock.schedule_once(self.set_colors)

	def set_colors(self, *args):
		back2 = self.get_color(self.background_state_color, 0)
		if back2 != None:
			self._background_color = back2

		Clock.schedule_once(lambda *a: self.on_state(self, self.state))
		return super().set_colors(self, *args)

	def on_state(self, widget, state):
		if state == "down":
			back = self.get_color(self.background_state_color, 1)
			if back != None:
				self._background_color = back
			
		elif state == "normal":
			back = self.get_color(self.background_state_color, 0)
			if back != None:
				self._background_color = back


class DraggableBoxLayout(ColoredBoxLayout, DraggableLayoutBehavior):
	def compare_pos_to_widget(self, widget, pos):
		if self.orientation == 'vertical':
			return 'before' if pos[1] >= widget.center_y else 'after'
		return 'before' if pos[0] < widget.center_x else 'after'






from kivy.compat import string_types
from kivy.factory import Factory
from kivy.properties import ListProperty, ObjectProperty, BooleanProperty
from kivygo.uix.button import ButtonEffect
from kivy.uix.dropdown import DropDown


class EffectSpinner(ButtonEffect):

    values = ListProperty([])
    '''Values that can be selected by the user. It must be a list of strings.
    '''

    text_autoupdate = BooleanProperty(False)
    '''Indicates if the spinner's :attr:`text` should be automatically
    updated with the first value of the :attr:`values` property.
    Setting it to True will cause the spinner to update its :attr:`text`
    property every time attr:`values` are changed.
    '''

    option_cls = ObjectProperty(ButtonEffect)
    '''Class used to display the options within the dropdown list displayed
    under the Spinner. The `text` property of the class will be used to
    represent the value.

    The option class requires:

    - a `text` property, used to display the value.
    - an `on_release` event, used to trigger the option when pressed/touched.
    - a :attr:`~kivy.uix.widget.Widget.size_hint_y` of None.
    - the :attr:`~kivy.uix.widget.Widget.height` to be set.
    '''

    dropdown_cls = ObjectProperty(DropDown)
    '''Class used to display the dropdown list when the Spinner is pressed.
    '''

    is_open = BooleanProperty(False)
    '''By default, the spinner is not open. Set to True to open it.
    '''

    sync_height = BooleanProperty(True)
    '''Each element in a dropdown list uses a default/user-supplied height.
    Set to True to propagate the Spinner's height value to each dropdown
    list element.
    '''

    def __init__(self, **kwargs):
        self._dropdown = None
        super().__init__(**kwargs)
        fbind = self.fbind
        build_dropdown = self._build_dropdown
        fbind('on_release', self._toggle_dropdown)
        fbind('dropdown_cls', build_dropdown)
        fbind('option_cls', build_dropdown)
        fbind('values', self._update_dropdown)
        fbind('size', self._update_dropdown_size)
        fbind('text_autoupdate', self._update_dropdown)
        build_dropdown()

    def _build_dropdown(self, *largs):
        if self._dropdown:
            self._dropdown.unbind(on_select=self._on_dropdown_select)
            self._dropdown.unbind(on_dismiss=self._close_dropdown)
            self._dropdown.dismiss()
            self._dropdown = None

        cls = self.dropdown_cls
        if isinstance(cls, string_types):
            cls = Factory.get(cls)
        self._dropdown = cls()
        self._dropdown.bind(on_select=self._on_dropdown_select)
        self._dropdown.bind(on_dismiss=self._close_dropdown)
        self._update_dropdown()

    def _update_dropdown_size(self, *largs):
        
        if not self.sync_height or not self._dropdown:
            return None

        container = self._dropdown.container
        if not container:
            return None
        
        h = self.height
        for item in container.children[:]:
            item.height = h

    def _update_dropdown(self, *largs):
        cls = self.option_cls
        values = self.values
        text_autoupdate = self.text_autoupdate
        if isinstance(cls, string_types):
            cls = Factory.get(cls)

        self._dropdown.clear_widgets()
        for value in values:
            item = cls(text=value, size_hint_y=None)
            item.height = self.height if self.sync_height else item.height
            item.bind(on_release=lambda option: self._dropdown.select(option.text))
            self._dropdown.add_widget(item)
        
        if text_autoupdate:
            if values:
                if not self.text or self.text not in values:
                    self.text = values[0]
            else:
                self.text = ''

    def _toggle_dropdown(self, *largs):
        if self.values:
            self.is_open = True

    def _close_dropdown(self, *largs):
        self.is_open = False

    def _on_dropdown_select(self, instance, data, *largs):
        self.text = data
        self.is_open = False

    def on_is_open(self, instance, value):
        if value:
                self._dropdown.open(self)
        else:
            self._dropdown.dismiss()



from kivy.lang import Builder
from kivy.uix.scrollview import ScrollView
from kivy.uix.codeinput import CodeInput
from kivy.uix.textinput import TextInput
from kivy.core.text.markup import MarkupLabel as CoreLabel
from kivy.properties import (
	ObjectProperty, StringProperty,
	BooleanProperty, ColorProperty,
	NumericProperty
)
from kivy.clock import Clock
from kivy.metrics import dp
from pygments.styles import get_style_by_name


Builder.load_string("""

<CodeEditor>:
	id: scroller
	code: text_code
	effect_cls: 'ScrollEffect'
	bar_width: '15dp'
	input_minimum_height: text_code.minimum_height
	TextCode:
		id: text_code
		size_hint: [None, None]
		on_text: scroller.update_width_code()
		height: max(self.minimum_height, scroller.height)
		font_size: root.font_size
		style: root.style
		background_color: [1, 1, 1, 0]
		text: root.text


<InputEditor>:
	id: scroller
	code: text_code
	effect_cls: 'ScrollEffect'
	bar_width: '15dp'
	input_minimum_height: text_code.minimum_height
	InputTerminal:
		id: text_code
		size_hint: None, None
		on_text: scroller.update_width_code()
		height: max(self.minimum_height, scroller.height)
		font_size: root.font_size
		style: root.style
		background_color: [1, 1, 1, 0]
		text: root.text
		foreground_color: root.text_color

		
""")


class InputTerminal(TextInput):

	can_scroll = True

	def on_touch_down(self, touch):
		if not self.can_scroll:
			# Não posso rolar
			return False
		return super().on_touch_down(touch)

	def on_touch_move(self, touch):
		if not self.can_scroll:
			# Não posso rolar
			return False
		return super().on_touch_move(touch)


class TextCode(CodeInput, InputTerminal):
	pass


class CodeEditorBase(ScrollView):
	
	code = ObjectProperty(None)
	bar_move = ''
	change_width = ObjectProperty(True)
	touch_pos = [0, 0]
	style = ObjectProperty(get_style_by_name('monokai'))
	text = StringProperty("")
	clickable = BooleanProperty(True)
	cursor_color = ColorProperty([1, 0, 0, 1])
	font_size = NumericProperty("12sp")
	text_color = ColorProperty([1, 1, 1, 1])
	input_minimum_height = NumericProperty(0)

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self.bind(size=self.update_width_code)
		self.bind(pos=self.update_width_code)
		Clock.schedule_once(self.start)

	
	def start(self, *args):
		self.code.cursor_color = self.cursor_color
		self.update_width_code()
		self.on_clickable()
	

	def on_clickable(self, *args):
		if not self.clickable:
			self.code.cursor_color = [0, 0, 0, 0]
			return None
		
		self.code.cursor_color = self.cursor_color


	# def on_touch_up(self, touch):
	#     if not self.clickable:
	#         return None
		
	#     return super().on_touch_up(touch)
	
	def on_touch_down(self, touch):
		self.touch_pos = touch.pos
		if not self.collide_point(*touch.pos):
			return False

		self.code.can_scroll = False
		if touch.x >= (self.x + self.width - self.bar_width):
			self.bar_move = 'left'
			return True
		elif touch.y <= (self.y + self.bar_width):
			self.bar_move = 'bottom'
			return True

		self.bar_move = ''
		self.code.can_scroll = True

		# if not self.clickable:
		#     return None
		
		return super().on_touch_down(touch)

	def maximum_text(self):
		label = CoreLabel(**self.code._get_line_options())
		lines = self.code.text.splitlines()
		max_width = 0
		for i in range(len(lines)):
			max_width = max(label.get_extents(lines[i])[0], max_width)
		return max_width + dp(20)

	def update_width_code(self, *args):
		if self.code == None or len(self.code._cursor) < 2:
			return None

		labels = self.code._lines_labels
		if not labels:
			return None
		
		if not self.change_width:
			self.code.width = self.width
			return None
		
		if self.code._cursor[1] >= len(labels):
			self.code._cursor = (0, 0)

		max_width = max(self.maximum_text(), self.width)
		if max_width != self.code.width:
			self.code.width = max_width
			self.scroll_x = 1
		

		current_label = labels[self.code._cursor[1]]
		if current_label.width > self.width:
			if self.scroll_x <= 0.5:
				self.scroll_x += 0.5
		else:
			self.scroll_x = 0
		
		if self.code.height <= self.height:
			self.scroll_y = 0
	
	def decrise_scroll(self, dt, direction):
		name = f'scroll_{direction}'
		n_scroll = getattr(self, name)
		
		if n_scroll >= dt:
			setattr(self, name, (n_scroll - dt))
		elif n_scroll > 0:
			setattr(self, name, 0)

	def incrise_scroll(self, dt, direction):
		name = f'scroll_{direction}'
		n_scroll = getattr(self, name)
		
		if n_scroll <= (1 - dt):
			setattr(self, name, (n_scroll + dt))
		elif n_scroll < 1:
			setattr(self, name, 1)

	def on_touch_move(self, touch):
		if self.code == None:
			return None
		
		if not self.collide_point(*touch.pos):
			return None
		
		if self.code._selection:
			# Desce ou sobe de acordo com a seleção
			x, y = touch.pos
			
			dt = (1 / self.code.height) * dp(40)
			if y < (self.y + dp(40)):
				self.decrise_scroll(dt, 'y')
			elif y > (self.y + self.height - dp(40)):
				self.incrise_scroll(dt, 'y')

			dt = (1 / self.code.width) * dp(40)
			if x < (self.x + dp(40)):
				self.decrise_scroll(dt, 'x')
			elif x > (self.x + self.width - dp(40)):
				self.incrise_scroll(dt, 'x')

		if not self.code.can_scroll:
			last_pos = self.touch_pos
			self.touch_pos = touch.pos

			cfd = (last_pos[1] - touch.y)
			distance = abs(cfd * (len(self.code._lines) * dp(17) / dp(185)) / 2.2)
			dt = round((1 / self.code.height * distance), 2)
			if self.bar_move == 'left':
				if cfd > dp(5):
					# Movimentar bar para baixo
					self.decrise_scroll(dt, 'y')
				elif cfd < dp(-5):
					# Movimentar bar para cima
					self.incrise_scroll(dt, 'y')
				else:
					self.touch_pos = last_pos
			
			cfd = (last_pos[0] - touch.x)
			dt = dp(7) * (1 / (self.code.width/max(abs(cfd), 1)))
			if self.bar_move == 'bottom':
				if cfd > dp(5):
					# Movimentar bar para esquerda
					self.decrise_scroll(dt, 'x')
				elif cfd < dp(-5):
					# Movimentar bar para a direita
					self.incrise_scroll(dt, 'x')
				else:
					self.touch_pos = last_pos
			return True

		# if not self.clickable:
		#     return None
	
		return super().on_touch_move(touch)


class InputEditor(CodeEditorBase):
	pass


class CodeEditor(CodeEditorBase):
	pass





from kivy.lang import Builder
from kivy.clock import Clock
from kivy.uix.label import Label
from kivy.uix.behaviors import ToggleButtonBehavior
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.anchorlayout import AnchorLayout
from kivy.uix.carousel import Carousel
from kivy.uix.widget import Widget
from kivy.uix.scrollview import ScrollView
from kivy.graphics import Rectangle
from kivy.utils import boundary
from kivy.properties import (
	ObjectProperty, NumericProperty,
	VariableListProperty, StringProperty,
	AliasProperty, BooleanProperty,
	BoundedNumericProperty, ListProperty
)

from kivygo.behaviors.touch_effecs import EffectBehavior
from kivygo.behaviors.hover import HoverBehavior


Builder.load_string('''

#:import DampedScrollEffect kivy.effects.dampedscroll.DampedScrollEffect

<AndroidTabsBar>:
	canvas.before: 
		Color:
			rgba: self._background_color
		RoundedRectangle:
			pos: self.pos
			size: self.size
			radius: self.radius
	canvas.after:
		Color:
			rgba: self.stroke_color
		Line:
			rounded_rectangle: [*self.pos, *self.size, *self.radius]
			width: self.stroke_width

<AndroidTabsLabel>:
	size_hint: None, 1
	halign: 'center'
	padding: '12dp', 0
	group: 'tabs'
	allow_no_selection: False
	text_color_normal: 1, 1, 1, .6
	text_color_active: 1, 1, 1, 1
	color: self.text_color_active if self.state is 'down' \
								else self.text_color_normal
	on_x: self._trigger_update_tab_indicator()
	on_width: self._trigger_update_tab_indicator()

<AndroidTabsScrollView>:
	size_hint: 1, 1
	do_scroll_y: False
	bar_color: 0, 0, 0, 0
	bar_inactive_color: 0, 0, 0, 0
	bar_width: 0
	effect_cls: DampedScrollEffect

<AndroidTabs>:
	carousel: carousel
	tab_bar: tab_bar
	anchor_y: 'top'

	AndroidTabsMain:
		padding: 0, tab_bar.height, 0, 0

		AndroidTabsCarousel:
			id: carousel
			anim_move_duration: root.anim_duration
			on_index: root.on_carousel_index(*args)
			on__offset: tab_bar.android_animation(*args)
			on_slides: self.index = root.default_tab
			on_slides: root.on_carousel_index(self, 0)

	AndroidTabsBar:
		id: tab_bar
		carousel: carousel
		scrollview: scrollview
		layout: layout
		size_hint: 1, None
		height: root.tab_bar_height

		AndroidTabsScrollView:
			id: scrollview
			on_width: tab_bar._trigger_update_tab_bar()

			GridLayout:
				id: layout
				rows: 1
				size_hint: None, 1
				width: self.minimum_width
				on_width: tab_bar._trigger_update_tab_bar()

				canvas.after:
					Color:
						rgba: root.tab_indicator_color
					Rectangle:
						pos: self.pos
						size: 0, root.tab_indicator_height
''')


class AndroidTabsException(Exception):
	pass


class AndroidTabsLabel(ToggleButtonBehavior, Label):

	text_color_normal = VariableListProperty([1, 1, 1, .6])
	'''Text color of the label when it is not selected.
	'''

	text_color_active = VariableListProperty([1])
	'''Text color of the label when it is selected.
	'''

	tab = ObjectProperty(None)
	tab_bar = ObjectProperty(None)

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self.min_space = 0

	def on_release(self):
		# if the label is selected load the relative tab from carousel
		if self.state == 'down':
			self.tab_bar.parent.carousel.load_slide(self.tab)

	def on_texture(self, widget, texture):
		# just save the minimum width of the label based of the content
		if not texture:
			return None
		
		self.width = texture.width
		self.min_space = self.width

	def _trigger_update_tab_indicator(self):
		# update the position and size of the indicator
		# when the label changes size or position
		if self.state == 'down':
			self.tab_bar.update_indicator(self.x, self.width)


class AndroidTabsBase(Widget):

	'''
	AndroidTabsBase allow you to create a tab.
	You must create a new class that inherits
	from AndroidTabsBase.
	In this way you have total control over the
	views of your tabbed panel.
	'''

	text = StringProperty('')
	'''It will be the label text of the tab.
	'''

	tab_label = ObjectProperty(None)
	'''It is the label object reference of the tab.
	'''

	def __init__(self, **kwargs):

		self.tab_label = AndroidTabsLabel(tab=self)
		super().__init__(**kwargs)

	def on_text(self, widget, text):
		# set the label text
		self.tab_label.text = self.text


class AndroidTabsMain(BoxLayout):
	'''
	AndroidTabsMain is just a boxlayout that contain
	the carousel. It allows you to have control over the carousel.
	'''
	pass


class AndroidTabsCarousel(Carousel):
	pass


class AndroidTabsScrollView(ScrollView):
	'''
	AndroidTabsScrollView hacked version to fix scroll_x manual setting.
	'''
	def goto(self, scroll_x, scroll_y):
		''' Update event value along with scroll_*
		'''
		def _update(e, x):
			if e:
				e.value = (e.max + e.min) * x

		if not (scroll_x == None):
			self.scroll_x = scroll_x
			_update(self.effect_x, scroll_x)

		if not (scroll_y == None):
			self.scroll_y = scroll_y
			_update(self.effect_y, scroll_y)


class AndroidTabsBar(BoxLayout, HoverBehavior, EffectBehavior):
	'''
	AndroidTabsBar is just a boxlayout that contain
	the scrollview for the tabs.
	It is also responsible to resize the tab label when it needed.
	'''

	stroke_color = ListProperty([0, 0 ,0 ,0])
	stroke_width = NumericProperty(2)

	background_color = ListProperty([0, 0, 0, 0])
	background_color_pos = ListProperty([0, 0, 0, 0])
	
	_background_color = ListProperty([0, 0, 0, 0])

	radius = ListProperty([0, 0, 0, 0])
	effect_color = ListProperty([0, 0, 0, 0])

	target = ObjectProperty(None, allownone=True)
	'''
	Is the carousel reference of the next tab / slide.
	When you go from "Tab A" to "Tab B", "Tab B" will be the
	target tab / slide of the carousel.
	'''

	def get_rect_instruction(self):

		for i in self.layout.canvas.after.children:
			if isinstance(i, Rectangle):
				return i

	indicator = AliasProperty(get_rect_instruction, cache=True)
	''' Is the Rectangle instruction reference of the tab indicator.
	'''

	def get_last_scroll_x(self):
		return self.scrollview.scroll_x

	last_scroll_x = AliasProperty(get_last_scroll_x, bind=('target', ), cache=True)
	'''
	Is the carousel reference of the next tab / slide.
	When you go from "Tab A" to "Tab B", "Tab B" will be the
	target tab / slide of the carousel.
	'''

	def __init__(self, **kwargs):
		self._trigger_update_tab_bar = Clock.schedule_once(self._update_tab_bar, 0)
		super().__init__(**kwargs)
		self.type_button = 'rounded'
		Clock.schedule_once(self.set_color)
	
	def set_color(self, *args):
		if isinstance(self.background_color[0], (int, float)):

			if self.background_color != [0, 0, 0, 0]:
				self._background_color = self.background_color

		elif self.background_color[0] != [0, 0, 0, 0]:
			self._background_color = self.background_color[0]
		
		if isinstance(self.background_color_pos[0], (int, float)):

			if self.background_color_pos != [0, 0, 0, 0]:
				self._background_color = self.background_color_pos
				
		elif self.background_color_pos[0] != [0, 0, 0, 0]:
			self._background_color = self.background_color_pos[0]

	def on_cursor_enter(self, *args):
		if not isinstance(self.background_color_pos[0], (int, float)):
			self._background_color = self.background_color_pos[1]

		return super().on_cursor_enter(*args)

	def on_cursor_leave(self, *args):
		if not isinstance(self.background_color_pos[0], (int, float)):
			self._background_color = self.background_color_pos[0]
			
		return super().on_cursor_leave(*args)

	def on_touch_down(self, touch):
		if not self.collide_point(*touch.pos):
			return False
		
		if len(self.background_color) == 2 and isinstance(self.background_color, (list, tuple)):
			self._background_color = self.background_color[1]
		
		self.ripple_show(touch)        
		return super().on_touch_down(touch)

	def on_touch_up(self, touch):
		if not self.collide_point(*touch.pos):
			return False
		
		if len(self.background_color) == 2 and isinstance(self.background_color, (list, tuple)):
			self._background_color = self.background_color[0]

		self.ripple_fade()
		return super().on_touch_up(touch)

	def _update_tab_bar(self, *args):
		# update width of the labels when it is needed
		width, tabs = self.scrollview.width, self.layout.children
		tabs_widths = [t.min_space for t in tabs if t.min_space]
		tabs_space = float(sum(tabs_widths))

		if not tabs_space:
			return None

		ratio = width / tabs_space
		use_ratio = True in (width / len(tabs) < w for w in tabs_widths)

		for t in tabs:

			t.width = t.min_space if tabs_space > width \
						else t.min_space * ratio if use_ratio is True \
						else width / len(tabs)

	def update_indicator(self, x, w):
		# update position and size of the indicator
		self.indicator.pos = (x, 0)
		self.indicator.size = (w, self.indicator.size[1])

	def tab_bar_autoscroll(self, target, step):
		# automatic scroll animation of the tab bar.
		bound_left = self.center_x
		bound_right = self.layout.width - bound_left
		dt = target.center_x - bound_left
		sx, sy = self.scrollview.convert_distance_to_scroll(dt, 0)

		# last scroll x of the tab bar
		lsx = self.last_scroll_x

		# determine scroll direction
		scroll_is_late = lsx < sx

		# distance to run
		dst = abs(lsx - sx) * step

		if not dst:
			return None

		if scroll_is_late and target.center_x > bound_left:
			x = lsx + dst

		elif not scroll_is_late and target.center_x < bound_right:
			x = lsx - dst
		
		x = boundary(x, 0.0, 1.0)
		self.scrollview.goto(x, None)


	def android_animation(self, carousel, offset):
		# try to reproduce the android animation effect.
		if offset != 0 and abs(offset) < carousel.width:
			forward = offset < 0
			offset = abs(offset)
			step = offset / float(carousel.width)
			distance = abs(offset - carousel.width)
			threshold = self.parent.anim_threshold
			breakpoint = carousel.width - (carousel.width * threshold)
			traveled = distance / breakpoint if breakpoint else 0
			break_step = 1.0 - traveled
			indicator_animation = self.parent.tab_indicator_anim

			skip_slide = carousel.slides[carousel._skip_slide] \
							if carousel._skip_slide != None else None
			next_slide = carousel.next_slide \
							if forward else carousel.previous_slide
			self.target = skip_slide if skip_slide else next_slide

			if not self.target:
				return None

			a = carousel.current_slide.tab_label
			b = self.target.tab_label

			self.tab_bar_autoscroll(b, step)

			if not indicator_animation:
				return None

			if step <= threshold:

				if forward:
					gap_w = abs((a.x + a.width) - (b.x + b.width))
					w_step = a.width + (gap_w * step)
					x_step = a.x

				else:
					gap = abs((a.x - b.x))
					x_step = a.x - gap * step
					w_step = a.width + gap * step

			else:

				if forward:
					x_step = a.x + abs((a.x - b.x)) * break_step
					gap_w = abs((a.x + a.width) - (b.x + b.width))
					ind_width = a.width + gap_w * threshold
					gap_w = ind_width - b.width
					w_step = ind_width - (gap_w * break_step)

				else:
					x_step = a.x - abs((a.x - b.x)) * threshold
					x_step = x_step - abs(x_step - b.x) * break_step
					ind_width = (a.x + a.width) - x_step if threshold else a.width
					gap_w = ind_width - b.width
					w_step = ind_width - (gap_w * break_step)
					w_step = w_step if w_step + x_step <= a.x + a.width else ind_width

			self.update_indicator(x_step, w_step)


class AndroidTabs(AnchorLayout):
	'''
	The AndroidTabs class.
	You can use it to create your own custom tabbed panel.
	'''

	default_tab = NumericProperty(0)
	'''Index of the default tab. Default to 0.
	'''

	tab_bar_height = NumericProperty('48dp')
	'''Height of the tab bar.
	'''

	tab_indicator_anim = BooleanProperty(True)
	'''
	Tab indicator animation. Default to True.
	If you do not want animation set it to False.
	'''

	tab_indicator_height = NumericProperty('2dp')
	'''Height of the tab indicator.
	'''

	tab_indicator_color = VariableListProperty([1])
	'''Color of the tab indicator.
	'''

	anim_duration = NumericProperty(0.2)
	'''Duration of the slide animation. Default to 0.2.
	'''

	anim_threshold = BoundedNumericProperty(0.8, min=0.0, max=1.0,
					errorhandler=lambda x: 0.0 if x < 0.0 else 1.0)
	'''
	Animation threshold allow you to change
	the tab indicator animation effect.
	Default to 0.8.
	'''

	def on_carousel_index(self, carousel, index):
		# when the index of the carousel change, update
		# tab indicator, select the current tab and reset threshold data.
		current_tab_label = carousel.current_slide.tab_label
		if current_tab_label.state == 'normal':
			current_tab_label._do_press()
		
		self.tab_bar.update_indicator(
			current_tab_label.x,
			current_tab_label.width)

	def add_widget(self, widget):
		# You can add only subclass of AndroidTabsBase.
		if len(self.children) >= 2:

			if not issubclass(widget.__class__, AndroidTabsBase):
				raise AndroidTabsException('AndroidTabs accept only subclass of AndroidTabsBase')

			widget.tab_label.tab_bar = self.tab_bar
			self.tab_bar.layout.add_widget(widget.tab_label)
			self.carousel.add_widget(widget)
			return None

		return super().add_widget(widget)

	def remove_widget(self, widget):
		# You can remove only subclass of AndroidTabsBase.
		if not issubclass(widget.__class__, AndroidTabsBase):

			raise AndroidTabsException('AndroidTabs can remove only subclass of AndroidTabBase')

		if widget.parent.parent == self.carousel:
			self.tab_bar.layout.remove_widget(widget.tab_label)
			self.carousel.remove_widget(widget)




from kivy.core.window import Window
from kivy.clock import Clock
from kivy.graphics import (
    RenderContext, Fbo, Color,
    ClearColor, ClearBuffers,
    Rectangle
)
from kivy.uix.floatlayout import FloatLayout
from kivy.properties import (
    StringProperty, ListProperty,
    ObjectProperty
)


header = '''

$HEADER$
uniform vec2 resolution;
uniform float time;
uniform vec2 mouse;

'''


shader_watter_bubble = (header + '''

void main(void)
{
	vec2 halfres = resolution.xy / 2.0;
	vec2 cpos = vec4(frag_modelview_mat * gl_FragCoord).xy;

	vec2 sinres = vec2(tan(resolution.x * time), -sin(resolution * time));

	cpos.x -= 0.5 * halfres.x * sin(time/2.0) + \
		0.3 * halfres.x * cos(time) + halfres.x;

	cpos.y -= 0.5 * halfres.y * sin(time/5.0) + \
		0.3 * halfres.y * sin(time) + halfres.y;

	float cLength = length(cpos);

	vec2 uv = \
		tex_coord0 + (cpos / cLength) * \
			sin(cLength / 50.0 - time * 2.0) / 15.0;

	vec3 col = texture2D(texture0, uv).xyz;
	gl_FragColor = vec4(col, 1.0);
}

''')




class ShaderWidget(FloatLayout):

    mouse_pos = ListProperty([100, 100])

    fs = StringProperty(None)

    texture = ObjectProperty(None)

    def __init__(self, **kwargs):

        Window.bind(mouse_pos=self.get_mouse_pos)

        self.canvas = RenderContext(
            use_parent_projection=True,
            use_parent_modelview=True,
            use_parent_frag_modelview=True
        )

        with self.canvas:
            self.fbo = Fbo(size=self.size)
            self.fbo_color = Color(1, 1, 1, 1)
            self.fbo_rect = Rectangle(size=self.size, pos=self.pos)

        with self.fbo:
            ClearColor(0, 0, 0, 0)
            ClearBuffers()

        super().__init__(**kwargs)

        self.fs = shader_watter_bubble
        Clock.schedule_interval(self.update_glsl, 0)

    def get_mouse_pos(self, w, pos):
        self.canvas['mouse'] = pos
        self.mouse_pos = pos

    def update_glsl(self, *largs):
        self.canvas['time'] = Clock.get_boottime()
        self.canvas['resolution'] = [float(v) for v in self.size]

    def on_fs(self, instance, value):

        shader = self.canvas.shader
        old_value = shader.fs
        shader.fs = value
        if not shader.success:
            shader.fs = old_value
            raise Exception('compilation failed')

    def add_widget(self, *args, **kwargs):
        c = self.canvas
        self.canvas = self.fbo
        super().add_widget(*args, **kwargs)
        self.canvas = c

    def remove_widget(self, *args, **kwargs):
        c = self.canvas
        self.canvas = self.fbo
        super().remove_widget(*args, **kwargs)
        self.canvas = c

    def on_size(self, instance, value):
        self.fbo.size = value
        self.texture = self.fbo.texture
        self.fbo_rect.size = value

    def on_pos(self, instance, value):
        self.fbo_rect.pos = value

    def on_texture(self, instance, value):
        self.fbo_rect.texture = value






from kivy.uix.widget import Widget
from kivy.clock import Clock
from kivy.graphics import (
    Color, Callback, Rotate,
    PushMatrix, PopMatrix,
    Translate, Quad
)

from kivy.graphics.opengl import (
    glBlendFunc, GL_SRC_ALPHA, GL_ONE, GL_ZERO,
    GL_SRC_COLOR, GL_ONE_MINUS_SRC_COLOR,
    GL_ONE_MINUS_SRC_ALPHA, GL_DST_ALPHA,
    GL_ONE_MINUS_DST_ALPHA, GL_DST_COLOR,
    GL_ONE_MINUS_DST_COLOR
)

from kivy.core.image import Image
from kivy.logger import Logger
from xml.dom.minidom import parse as parse_xml
from kivy.properties import (
    NumericProperty, BooleanProperty,
    ListProperty, StringProperty,
    ObjectProperty
)

import os, sys
import math, random


def random_variance(base, variance):
    return base + variance * (random.random() * 2.0 - 1.0)

def random_color_variance(base, variance):
    return [min(max(0.0, (random_variance(base[i], variance[i]))), 1.0) for i in range(4)]


EMITTER_TYPE_GRAVITY = 0
EMITTER_TYPE_RADIAL = 1

BLEND_FUNC = {0: GL_ZERO,
            1: GL_ONE,
            0x300: GL_SRC_COLOR,
            0x301: GL_ONE_MINUS_SRC_COLOR,
            0x302: GL_SRC_ALPHA,
            0x303: GL_ONE_MINUS_SRC_ALPHA,
            0x304: GL_DST_ALPHA,
            0x305: GL_ONE_MINUS_DST_ALPHA,
            0x306: GL_DST_COLOR,
            0x307: GL_ONE_MINUS_DST_COLOR
}


class Particle(object):
    x, y, rotation, current_time = -256, -256, 0, 0
    scale, total_time = 1.0, 0.
    color = [1.0, 1.0, 1.0, 1.0]
    color_delta = [0.0, 0.0, 0.0, 0.0]
    start_x, start_y, velocity_x, velocity_y = 0, 0, 0, 0
    radial_acceleration, tangent_acceleration = 0, 0
    emit_radius, emit_radius_delta = 0, 0
    emit_rotation, emit_rotation_delta = 0, 0
    rotation_delta, scale_delta = 0, 0


class ParticleSystem(Widget):
    max_num_particles = NumericProperty(200)
    life_span = NumericProperty(2)
    texture = ObjectProperty(None)
    texture_path = StringProperty(None)
    life_span_variance = NumericProperty(0)
    start_size = NumericProperty(16)
    start_size_variance = NumericProperty(0)
    end_size = NumericProperty(16)
    end_size_variance = NumericProperty(0)
    emit_angle = NumericProperty(0)
    emit_angle_variance = NumericProperty(0)
    start_rotation = NumericProperty(0)
    start_rotation_variance = NumericProperty(0)
    end_rotation = NumericProperty(0)
    end_rotation_variance = NumericProperty(0)
    emitter_x_variance = NumericProperty(100)
    emitter_y_variance = NumericProperty(100)
    gravity_x = NumericProperty(0)
    gravity_y = NumericProperty(0)
    speed = NumericProperty(0)
    speed_variance = NumericProperty(0)
    radial_acceleration = NumericProperty(100)
    radial_acceleration_variance = NumericProperty(0)
    tangential_acceleration = NumericProperty(0)
    tangential_acceleration_variance = NumericProperty(0)
    max_radius = NumericProperty(100)
    max_radius_variance = NumericProperty(0)
    min_radius = NumericProperty(50)
    rotate_per_second = NumericProperty(0)
    rotate_per_second_variance = NumericProperty(0)
    start_color = ListProperty([1., 1., 1., 1.])
    start_color_variance = ListProperty([1., 1., 1., 1.])
    end_color = ListProperty([1., 1., 1., 1.])
    end_color_variance = ListProperty([1., 1., 1., 1.])
    blend_factor_source = NumericProperty(770)
    blend_factor_dest = NumericProperty(1)
    emitter_type = NumericProperty(0)

    update_interval = NumericProperty(1. / 30.)
    _is_paused = BooleanProperty(False)

    def __init__(self, config, **kwargs):
        super().__init__(**kwargs)
        self.capacity = 0
        self.particles = list()
        self.particles_dict = dict()
        self.emission_time = 0.0
        self.frame_time = 0.0
        self.num_particles = 0

        if config != None:
            self._parse_config(config)
        self.emission_rate = self.max_num_particles / self.life_span
        self.initial_capacity = self.max_num_particles
        self.max_capacity = self.max_num_particles
        self._raise_capacity(self.initial_capacity)

        with self.canvas.before:
            Callback(self._set_blend_func)
        with self.canvas.after:
            Callback(self._reset_blend_func)

        Clock.schedule_once(self._update, self.update_interval)

    def start(self, duration=sys.maxsize):
        if self.emission_rate != 0:
            self.emission_time = duration

    def stop(self, clear=False):
        self.emission_time = 0.0
        if clear:
            self.num_particles = 0
            self.particles_dict = dict()
            self.canvas.clear()

    def on_max_num_particles(self, instance, value):
        self.max_capacity = value
        if self.capacity < value:
            self._raise_capacity(self.max_capacity - self.capacity)
        elif self.capacity > value:
            self._lower_capacity(self.capacity - self.max_capacity)
        self.emission_rate = self.max_num_particles / self.life_span

    def on_texture(self, instance, value):
        for p in self.particles:
            try:
                self.particles_dict[p]['rect'].texture = self.texture
            except KeyError:
                # if particle isn't initialized yet, you can't change its texture.
                pass

    def on_life_span(self, instance, value):
        self.emission_rate = self.max_num_particles / value

    def _set_blend_func(self, instruction):
        glBlendFunc(GL_SRC_ALPHA, GL_ONE)

    def _reset_blend_func(self, instruction):
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

    def _parse_config(self, config):
        self._config = parse_xml(config)

        texture_path = self._parse_data('texture', 'name')
        config_dir_path = os.path.dirname(os.path.abspath(config))
        path = os.path.join(config_dir_path, texture_path)
        if os.path.exists(path):
            self.texture_path = path
        else:
            self.texture_path = texture_path

        self.texture = Image(self.texture_path).texture
        self.emitter_x = float(self._parse_data('sourcePosition', 'x'))
        self.emitter_y = float(self._parse_data('sourcePosition', 'y'))
        self.emitter_x_variance = float(self._parse_data('sourcePositionVariance', 'x'))
        self.emitter_y_variance = float(self._parse_data('sourcePositionVariance', 'y'))
        self.gravity_x = float(self._parse_data('gravity', 'x'))
        self.gravity_y = float(self._parse_data('gravity', 'y'))
        self.emitter_type = int(self._parse_data('emitterType'))
        self.max_num_particles = int(self._parse_data('maxParticles'))
        self.life_span = max(0.01, float(self._parse_data('particleLifeSpan')))
        self.life_span_variance = float(self._parse_data('particleLifespanVariance'))
        self.start_size = float(self._parse_data('startParticleSize'))
        self.start_size_variance = float(self._parse_data('startParticleSizeVariance'))
        self.end_size = float(self._parse_data('finishParticleSize'))
        self.end_size_variance = float(self._parse_data('FinishParticleSizeVariance'))
        self.emit_angle = math.radians(float(self._parse_data('angle')))
        self.emit_angle_variance = math.radians(float(self._parse_data('angleVariance')))
        self.start_rotation = math.radians(float(self._parse_data('rotationStart')))
        self.start_rotation_variance = math.radians(float(self._parse_data('rotationStartVariance')))
        self.end_rotation = math.radians(float(self._parse_data('rotationEnd')))
        self.end_rotation_variance = math.radians(float(self._parse_data('rotationEndVariance')))
        self.speed = float(self._parse_data('speed'))
        self.speed_variance = float(self._parse_data('speedVariance'))
        self.radial_acceleration = float(self._parse_data('radialAcceleration'))
        self.radial_acceleration_variance = float(self._parse_data('radialAccelVariance'))
        self.tangential_acceleration = float(self._parse_data('tangentialAcceleration'))
        self.tangential_acceleration_variance = float(self._parse_data('tangentialAccelVariance'))
        self.max_radius = float(self._parse_data('maxRadius'))
        self.max_radius_variance = float(self._parse_data('maxRadiusVariance'))
        self.min_radius = float(self._parse_data('minRadius'))
        self.rotate_per_second = math.radians(float(self._parse_data('rotatePerSecond')))
        self.rotate_per_second_variance = math.radians(float(self._parse_data('rotatePerSecondVariance')))
        self.start_color = self._parse_color('startColor')
        self.start_color_variance = self._parse_color('startColorVariance')
        self.end_color = self._parse_color('finishColor')
        self.end_color_variance = self._parse_color('finishColorVariance')
        self.blend_factor_source = self._parse_blend('blendFuncSource')
        self.blend_factor_dest = self._parse_blend('blendFuncDestination')

    def _parse_data(self, name, attribute='value'):
        return self._config.getElementsByTagName(name)[0].getAttribute(attribute)

    def _parse_color(self, name):
        return [float(self._parse_data(name, 'red')), float(self._parse_data(name, 'green')), float(self._parse_data(name, 'blue')), float(self._parse_data(name, 'alpha'))]

    def _parse_blend(self, name):
        value = int(self._parse_data(name))
        return BLEND_FUNC[value]

    def pause(self):
        self._is_paused = True

    def resume(self):
        self._is_paused = False
        Clock.schedule_once(self._update, self.update_interval)

    def _update(self, dt):
        self._advance_time(dt)
        self._render()
        if not self._is_paused:
            Clock.schedule_once(self._update, self.update_interval)

    def _create_particle(self):
        return Particle()

    def _init_particle(self, particle):
        life_span = random_variance(self.life_span, self.life_span_variance)
        if life_span <= 0.0:
            return

        particle.current_time = 0.0
        particle.total_time = life_span

        particle.x = random_variance(self.emitter_x, self.emitter_x_variance)
        particle.y = random_variance(self.emitter_y, self.emitter_y_variance)
        particle.start_x = self.emitter_x
        particle.start_y = self.emitter_y

        angle = random_variance(self.emit_angle, self.emit_angle_variance)
        speed = random_variance(self.speed, self.speed_variance)
        particle.velocity_x = speed * math.cos(angle)
        particle.velocity_y = speed * math.sin(angle)

        particle.emit_radius = random_variance(self.max_radius, self.max_radius_variance)
        particle.emit_radius_delta = (self.max_radius - self.min_radius) / life_span

        particle.emit_rotation = random_variance(self.emit_angle, self.emit_angle_variance)
        particle.emit_rotation_delta = random_variance(self.rotate_per_second, self.rotate_per_second_variance)

        particle.radial_acceleration = random_variance(self.radial_acceleration, self.radial_acceleration_variance)
        particle.tangent_acceleration = random_variance(self.tangential_acceleration, self.tangential_acceleration_variance)

        start_size = random_variance(self.start_size, self.start_size_variance)
        end_size = random_variance(self.end_size, self.end_size_variance)

        start_size = max(0.1, start_size)
        end_size = max(0.1, end_size)

        particle.scale = start_size / self.texture.width
        particle.scale_delta = ((end_size - start_size) / life_span) / self.texture.width

        # colors
        start_color = random_color_variance(self.start_color, self.start_color_variance)
        end_color = random_color_variance(self.end_color, self.end_color_variance)

        particle.color_delta = [(end_color[i] - start_color[i]) / life_span for i in range(4)]
        particle.color = start_color

        # rotation
        start_rotation = random_variance(self.start_rotation, self.start_rotation_variance)
        end_rotation = random_variance(self.end_rotation, self.end_rotation_variance)
        particle.rotation = start_rotation
        particle.rotation_delta = (end_rotation - start_rotation) / life_span

    def _advance_particle(self, particle, passed_time):
        passed_time = min(passed_time, particle.total_time - particle.current_time)
        particle.current_time += passed_time

        if self.emitter_type == EMITTER_TYPE_RADIAL:
            particle.emit_rotation += particle.emit_rotation_delta * passed_time
            particle.emit_radius -= particle.emit_radius_delta * passed_time
            particle.x = self.emitter_x - math.cos(particle.emit_rotation) * particle.emit_radius
            particle.y = self.emitter_y - math.sin(particle.emit_rotation) * particle.emit_radius

            if particle.emit_radius < self.min_radius:
                particle.current_time = particle.total_time

        else:
            distance_x = particle.x - particle.start_x
            distance_y = particle.y - particle.start_y
            distance_scalar = math.sqrt(distance_x * distance_x + distance_y * distance_y)
            if distance_scalar < 0.01:
                distance_scalar = 0.01

            radial_x = distance_x / distance_scalar
            radial_y = distance_y / distance_scalar
            tangential_x = radial_x
            tangential_y = radial_y

            radial_x *= particle.radial_acceleration
            radial_y *= particle.radial_acceleration

            new_y = tangential_x
            tangential_x = -tangential_y * particle.tangent_acceleration
            tangential_y = new_y * particle.tangent_acceleration

            particle.velocity_x += passed_time * (self.gravity_x + radial_x + tangential_x)
            particle.velocity_y += passed_time * (self.gravity_y + radial_y + tangential_y)

            particle.x += particle.velocity_x * passed_time
            particle.y += particle.velocity_y * passed_time

        particle.scale += particle.scale_delta * passed_time
        particle.rotation += particle.rotation_delta * passed_time

        particle.color = [particle.color[i] + particle.color_delta[i] * passed_time for i in range(4)]

    def _raise_capacity(self, by_amount):
        old_capacity = self.capacity
        new_capacity = min(self.max_capacity, self.capacity + by_amount)

        for _ in range(int(new_capacity - old_capacity)):
            self.particles.append(self._create_particle())

        self.num_particles = int(new_capacity)
        self.capacity = new_capacity

    def _lower_capacity(self, by_amount):
        old_capacity = self.capacity
        new_capacity = max(0, self.capacity - by_amount)

        for i in range(int(old_capacity - new_capacity)):
            try:
                self.canvas.remove(self.particles_dict[self.particles.pop()]['rect'])
            except:
                pass

        self.num_particles = int(new_capacity)
        self.capacity = new_capacity

    def _advance_time(self, passed_time):
        particle_index = 0

        # advance existing particles
        while particle_index < self.num_particles:
            particle = self.particles[particle_index]
            if particle.current_time < particle.total_time:
                self._advance_particle(particle, passed_time)
                particle_index += 1
            else:
                if particle_index != self.num_particles - 1:
                    next_particle = self.particles[self.num_particles - 1]
                    self.particles[self.num_particles - 1] = particle
                    self.particles[particle_index] = next_particle
                self.num_particles -= 1
                if self.num_particles == 0:
                    Logger.debug('Particle: COMPLETE')

        # create and advance new particles
        if self.emission_time > 0:
            time_between_particles = 1.0 / self.emission_rate
            self.frame_time += passed_time

            while self.frame_time > 0:
                if self.num_particles < self.max_capacity:
                    if self.num_particles == self.capacity:
                        self._raise_capacity(self.capacity)

                    particle = self.particles[self.num_particles]
                    self.num_particles += 1
                    self._init_particle(particle)
                    self._advance_particle(particle, self.frame_time)

                self.frame_time -= time_between_particles

            if self.emission_time != sys.maxsize:
                self.emission_time = max(0.0, self.emission_time - passed_time)

    def _render(self):
        if self.num_particles == 0:
            return None
        
        for i in range(self.num_particles):
            particle = self.particles[i]
            size = (self.texture.size[0] * particle.scale, self.texture.size[1] * particle.scale)
            if particle not in self.particles_dict:
                self.particles_dict[particle] = dict()
                color = particle.color[:]
                with self.canvas:
                    self.particles_dict[particle]['color'] = Color(color[0], color[1], color[2], color[3])
                    PushMatrix()
                    self.particles_dict[particle]['translate'] = Translate()
                    self.particles_dict[particle]['rotate'] = Rotate()
                    self.particles_dict[particle]['rotate'].set(particle.rotation, 0, 0, 1)
                    self.particles_dict[particle]['rect'] = Quad(texture=self.texture, points=(-size[0] * 0.5, -size[1] * 0.5, size[0] * 0.5, -size[1] * 0.5, size[0] * 0.5,  size[1] * 0.5, -size[0] * 0.5,  size[1] * 0.5))
                    self.particles_dict[particle]['translate'].xy = (particle.x, particle.y)
                    PopMatrix()
            else:
                self.particles_dict[particle]['rotate'].angle = particle.rotation
                self.particles_dict[particle]['translate'].xy = (particle.x, particle.y)
                self.particles_dict[particle]['color'].rgba = particle.color
                self.particles_dict[particle]['rect'].points = (-size[0] * 0.5, -size[1] * 0.5, size[0] * 0.5, -size[1] * 0.5, size[0] * 0.5,  size[1] * 0.5, -size[0] * 0.5,  size[1] * 0.5)




from kivy.properties import ObjectProperty, ListProperty, NumericProperty
from kivy.utils import get_color_from_hex
from kivy.lang.builder import Builder
from kivy.core.window import Window
from kivy.clock import Clock
from kivy.metrics import dp

from kivy.uix.scrollview import ScrollView
from kivy.uix.anchorlayout import AnchorLayout


Builder.load_string("""

#:import hex kivy.utils.get_color_from_hex


<BarScroll>:
	size_hint_x: None
	width: '40dp'
	canvas.after:
		Color:
			rgba: self._background_color
		RoundedRectangle:
			size: [ ( self.bar_width * 2 ), self.height ]
			pos: [ ( self.x + ( self.width / 2 ) - (self.bar_width * 2 / 2) ), self.y ]
			radius: self.radius
		Color:
			rgba: self._bar_color
		RoundedRectangle:
			size: [self.bar_width, self.bar_height]
			pos: [ ( self.x + ( self.width / 2 ) - (self.bar_width / 2 ) ), self.bar_y ]
			radius: self.bar_radius

<ScrollViewBar>:
	bar_color: [0, 0, 0, 0]
	bar_inactive_color: [0, 0, 0, 0]
	effect_cls: 'ScrollEffect'

	
""")


class ScrollViewBar(ScrollView):
	pass


class BarScroll(AnchorLayout):

	scroll_view = ObjectProperty(None)
	_last_scroll_view = None
	bar_width = NumericProperty(dp(4))
	bar_height = NumericProperty(0)
	bar_y = NumericProperty(0)
	
	bar_color = ListProperty(get_color_from_hex('737373'))
	_bar_color = ListProperty([0, 0, 0, 0])
	
	background_color = ListProperty(get_color_from_hex('dfe0e5'))
	_background_color = ListProperty([0, 0, 0, 0])

	radius = ListProperty([0, 0, 0, 0])
	bar_radius = ListProperty([0, 0, 0, 0])
	can_scroll = False

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		Clock.schedule_once(self.start)
	
	def start(self, *args):
		Window.bind(on_restore=self.update_scroll)
		self.bind(size=self.update_scroll)
		self.bind(pos=self.update_scroll)
		
		if self.scroll_view != None:
			self._last_scroll_view = self.scroll_view
			self.scroll_view.bind(vbar=self.update_scroll)
		
		Clock.schedule_interval(self.update_scroll, 0.001)


	def on_scroll_view(self, *args):
		if self.scroll_view == None:
			self._background_color = [0, 0, 0, 0]
			self._bar_color = [0, 0, 0, 0]
			return None
		
		if self._last_scroll_view != None:
			self._last_scroll_view.unbind(vbar=self.update_scroll)

		self._background_color = self.background_color
		self._last_scroll_view = self.scroll_view
		self.scroll_view.bind(vbar=self.update_scroll)
		Clock.schedule_once(self.update_scroll)

	def update_scroll(self, *args):
		if Clock.get_rfps() > 0:
			Clock.unschedule(self.update_scroll)
		
		if self.scroll_view == None or not self.scroll_view.children:
			self._bar_color = [0, 0, 0, 0]
			return None
		
		if self.scroll_view.height >= self.scroll_view.children[0].height:
			self._bar_color = [0, 0, 0, 0]
		else:
			self._bar_color = self.bar_color

		vy, vh = self.scroll_view._get_vbar()
		self.bar_y = self.y + (self.height * vy)
		self.bar_height = (self.height * vh)
	
	def on_touch_down(self, touch):
		if self.collide_point(*touch.pos):
			self.can_scroll = True
		
		return super().on_touch_down(touch)

	def on_touch_up(self, touch):
		if self.can_scroll:
			self.can_scroll = False
		
		return super().on_touch_up(touch)

	def on_touch_move(self, touch):
		if not self.can_scroll or self.scroll_view == None:
			return None
			
		height = self.scroll_view.height
		dy = touch.dy / float(max((height - (height * self.scroll_view.vbar[1])), 1))
		self.scroll_view.scroll_y = float(min(max(self.scroll_view.scroll_y + dy, 0), 1))





from kivy.graphics import (
    Line as KivyLine,
    Color, Mesh,
)

from kivy.graphics.tesselator import (
    Tesselator, WINDING_ODD,
    TYPE_POLYGONS,
)

from svg.path.path import (
    Line, CubicBezier,
    Close, Move,
)

from kivygo.utils import (
    line_points, bezier_points,
    get_all_points, parse_svg,
    find_center,
)

from kivygo.animation import Animation
from svg.path import parse_path
from collections import OrderedDict


class Kivg():
    def __init__(self, widget, *args):
        """
        widget: widget to draw svg upon
        """
        self.b = widget
        self._fill = True  # Fill path with color after drawing
        self._LINE_WIDTH = 2
        self._LINE_COLOR = [0, 0, 0, 1]
        self._DUR = 0.02
        self.psf = ""  # Previous svg file - Don't re-find path for same file in a row

    def get_tess(self, shapes):
        tess = Tesselator()
        for shape in shapes:
            if len(shape) >= 3:
                tess.add_contour(shape)
        return tess

    def get_mesh(self, shapes):
        tess = self.get_tess(shapes)
        ret = tess.tesselate(WINDING_ODD, TYPE_POLYGONS)
        return tess.meshes

    def fill_up(self, shapes, color):
        meshes = self.get_mesh(shapes)
        with self.b.canvas:
            Color(*color[:3], getattr(self.b, "mesh_opacity"))
            for vertices, indices in meshes:
                Mesh(vertices=vertices, indices=indices, mode="triangle_fan")

    def fill_up_shapes(self, *args):
        for id_, closed_paths in self.closed_shapes.items():
            c = self.closed_shapes[id_]["color"]
            self.fill_up(closed_paths[id_ + "shapes"], c)

    def fill_up_shapes_anim(self, shapes, *args):
        for shape in shapes:
            c = shape[0]
            self.fill_up([shape[1]], c)

    def anim_on_comp(self, *args):
        self.curr_count += 1
        self.prev_shapes.append(self.curr_shape)
        if self.curr_count < len(self.all_anim):
            id_, a = self.all_anim[self.curr_count]
            setattr(self, "curr_id", id_)
            setattr(self, "curr_clr", self.closed_shapes[id_]["color"])
            a.bind(on_progress=self.track_progress)
            a.start(self.b)

    def shape_animate(self, svg_file, anim_config_list=[], on_complete=None):
        """
        svg_file: svg file name

        anim_config_list: a list of dicts with keys id_ to animate and from_, the direction of animation

        on_complete: optional function to call after total completion
        """
        self.draw(svg_file, from_shape_anim=True)
        setattr(self.b, "mesh_opacity", 1)

        self.all_anim = []
        self.curr_count = 0
        for i, config in enumerate(anim_config_list):
            anim_list = self._shape_animate(
                config["id_"],
                config.get("from_", None),
                config.get("t", "out_sine"),
                config.get("d", .3),
            )

            if anim_list:
                anim = anim_list[0]
                for a in anim_list[1:]:
                    anim &= a

                anim.bind(on_complete=self.anim_on_comp)
                self.all_anim.append((config["id_"], anim))
            else:
                setattr(self, "curr_id", config["id_"])
                setattr(self, "curr_clr",
                        self.closed_shapes[config["id_"]]["color"])
                self.track_progress()

        id_, a = self.all_anim[0]
        setattr(self, "curr_id", id_)
        setattr(self, "curr_clr", self.closed_shapes[id_]["color"])

        a.cancel_all(self.b)
        a.bind(on_progress=self.track_progress)

        if on_complete:
            self.all_anim[-1][1].bind(on_complete=on_complete)

        a.start(self.b)

    def _shape_animate(self, id_, from_, t, d):
        line_count = 0
        bezier_count = 0
        anim_list = []
        self.prev_shapes = []
        self.curr_shape = []
        if self.closed_shapes.get(id_, None):
            tmp = []
            for s in self.closed_shapes[id_][id_ + "paths"]:
                tmp2 = []
                for e in s:
                    if isinstance(e, Line):
                        lp = line_points(
                            e, *self.b.size, *self.b.pos,
                            *self.sw_size, self.sf
                        )

                        tmp2.append([
                            (lp[0], lp[1]),
                            (lp[2], lp[3])
                        ])

                    if isinstance(e, CubicBezier):
                        bp = bezier_points(
                            e, *self.b.size, *self.b.pos,
                            *self.sw_size, self.sf
                        )

                        tmp2.append([
                            (bp[0], bp[1]),
                            (bp[2], bp[3]),
                            (bp[4], bp[5]),
                            (bp[6], bp[7])
                        ])

                tmp.append(tmp2)

            l = []
            for each in tmp:
                for e in each:
                    for i in e:
                        if from_ in ("left", "right", "center_x"):
                            l.append(i[0])
                        else:
                            l.append(i[1])

            base_point = 0
            if from_ in ("top", "right"):
                # rightmost/topmost point to start animation from
                base_point = max(l)
            elif from_ in ("left", "bottom"):
                # leftmost/bottommost point to start animation from
                base_point = min(l)
            elif from_ in ("center_x", "center_y"):
                base_point = find_center(sorted(l))

            for each in tmp:
                for e in each:
                    # Line
                    if len(e) == 2:
                        nm_mesh = f"{id_}_mesh_line{line_count}"
                        value_sx = value_sy = base_point
                        value_ex = value_ey = base_point

                        if from_ in ("left", "right", "center_x"):
                            value_sx = e[0][0]
                            value_ex = e[1][0]

                            anim_list.append(
                                Animation(
                                    d=d, t=t,
                                    **{
                                        f"{nm_mesh}_start_x": e[0][0],
                                        f"{nm_mesh}_end_x": e[1][0]
                                    }
                                )
                            )

                        if from_ in ("top", "bottom", "center_y"):
                            value_sy = e[0][1]
                            value_ey = e[1][1]

                            anim_list.append(
                                Animation(
                                    d=d, t=t,
                                    **{
                                        f"{nm_mesh}_start_y": e[0][1],
                                        f"{nm_mesh}_end_y": e[1][1]
                                    }
                                )
                            )

                        setattr(self.b, f"{nm_mesh}_start_x", value_sx)
                        setattr(self.b, f"{nm_mesh}_start_y", value_sy)
                        setattr(self.b, f"{nm_mesh}_end_x", value_ex)
                        setattr(self.b, f"{nm_mesh}_end_y", value_ey)
                        line_count += 1

                    # Bezier
                    if len(e) == 4:
                        nm_mesh = f"{id_}_mesh_bezier{bezier_count}"

                        value_sx = value_sy = base_point
                        value_ex = value_ey = base_point
                        value_ctrl_1x = value_ctrl_1y = base_point
                        value_ctrl_2x = value_ctrl_2y = base_point

                        if from_ in ("left", "right", "center_x"):
                            value_sx = e[0][0]
                            value_ctrl_1x = e[1][0]
                            value_ctrl_2x = e[2][0]
                            value_ex = e[3][0]

                            anim_list.append(
                                Animation(
                                    d=d, t=t,
                                    **{
                                        f"{nm_mesh}_start_x": e[0][0],
                                        f"{nm_mesh}_control1_x": e[1][0],
                                        f"{nm_mesh}_control2_x": e[2][0],
                                        f"{nm_mesh}_end_x": e[3][0]
                                    }
                                )
                            )

                        if from_ in ("top", "bottom", "center_y"):
                            value_sy = e[0][1]
                            value_ctrl_1y = e[1][1]
                            value_ctrl_2y = e[2][1]
                            value_ey = e[3][1]

                            anim_list.append(
                                Animation(
                                    d=d, t=t,
                                    **{
                                        f"{nm_mesh}_start_y": e[0][1],
                                        f"{nm_mesh}_control1_y": e[1][1],
                                        f"{nm_mesh}_control2_y": e[2][1],
                                        f"{nm_mesh}_end_y": e[3][1]
                                    }
                                )
                            )

                        setattr(self.b, f"{nm_mesh}_start_x", value_sx)
                        setattr(self.b, f"{nm_mesh}_start_y", value_sy)
                        setattr(self.b, f"{nm_mesh}_control1_x", value_ctrl_1x)
                        setattr(self.b, f"{nm_mesh}_control1_y", value_ctrl_1y)
                        setattr(self.b, f"{nm_mesh}_control2_x", value_ctrl_2x)
                        setattr(self.b, f"{nm_mesh}_control2_y", value_ctrl_2y)
                        setattr(self.b, f"{nm_mesh}_end_x", value_ex)
                        setattr(self.b, f"{nm_mesh}_end_y", value_ey)
                        bezier_count += 1

            setattr(self, f"{id_}_tmp", tmp)
            return anim_list

    def track_progress(self, *args):

        id_ = getattr(self, "curr_id")
        shape_list = []
        line_count = 0
        bezier_count = 0

        for each in getattr(self, "{}_tmp".format(id_)):
            for e in each:
                # Line
                if len(e) == 2:
                    nm_mesh = f"{id_}_mesh_line{line_count}"

                    shape_list.extend([
                        getattr(self.b, f"{nm_mesh}_start_x"),
                        getattr(self.b, f"{nm_mesh}_start_y"),
                        getattr(self.b, f"{nm_mesh}_end_x"),
                        getattr(self.b, f"{nm_mesh}_end_y")
                    ])
                    line_count += 1

                # Bezier
                if len(e) == 4:
                    nm_mesh = f"{id_}_mesh_bezier{bezier_count}"
                    shape_list.extend(
                        get_all_points(
                            (getattr(self.b, f"{nm_mesh}_start_x"), getattr(
                                self.b, f"{nm_mesh}_start_y")),
                            (getattr(self.b, f"{nm_mesh}_control1_x"), getattr(
                                self.b, f"{nm_mesh}_control1_y")),
                            (getattr(self.b, f"{nm_mesh}_control2_x"), getattr(
                                self.b, f"{nm_mesh}_control2_y")),
                            (getattr(self.b, f"{nm_mesh}_end_x"), getattr(
                                self.b, f"{nm_mesh}_end_y"))
                        )
                    )
                    bezier_count += 1

        self.b.canvas.clear()
        self.curr_shape = (getattr(self, "curr_clr"), shape_list)
        s = [*self.prev_shapes, self.curr_shape]
        self.fill_up_shapes_anim(s)

    def draw(self, svg_file, animate=False, anim_type="seq", *args, **kwargs):
        """
        Function to animate

        Call this function with an svg file name to animate that svg

        ------------
        Extra arguments:

        fill: Whether to fill the drawing at the end using same png, defaults true,
         unexpected result if png with same name is not available

        outline_width: Line width for drawing, default 2

        line_color: Line Color for drawing, default [0,0,0,1]

        dur: Duration of each small path drawing animation, default .02

        anim_type: "seq"/"par" for sequence/parallel
        """
        self.fill = kwargs.get("fill", self._fill)
        self.outline_width = kwargs.get("outline_width", self._LINE_WIDTH)
        self.LINE_COLOR = kwargs.get("line_color", self._LINE_COLOR)
        self.DUR = kwargs.get("dur", self._DUR)
        from_shape_anim = kwargs.get("from_shape_anim", False)
        anim_type = (anim_type) if anim_type in {"seq", "par"} else ("seq")

        self.sf = svg_file
        if self.sf != self.psf:
            self.sw_size, path_strings = parse_svg(svg_file)

            self.path = []
            self.closed_shapes = OrderedDict()

            for path_string, id_, clr in path_strings:

                move_found = False
                tmp = []
                self.closed_shapes[id_] = dict()
                self.closed_shapes[id_][f"{id_}paths"] = []

                # for drawing meshes
                self.closed_shapes[id_][f"{id_}shapes"] = []
                self.closed_shapes[id_]["color"] = clr
                _path = parse_path(path_string)

                for e in _path:
                    self.path.append(e)

                    if isinstance(e, Close) or (isinstance(e, Move) and move_found):
                        self.closed_shapes[id_][f"{id_}paths"].append(tmp)
                        move_found = False

                    # shape started
                    if isinstance(e, Move):
                        tmp = []
                        move_found = True

                    if not isinstance(e, Move) and move_found:
                        tmp.append(e)

            self.psf = self.sf

        anim_list = self._calc_paths(animate)
        if not from_shape_anim:
            if animate:
                anim = anim_list[0]
                for a in anim_list[1:]:
                    if anim_type == "seq":
                        anim += a
                    else:
                        anim &= a

            if self.fill:
                setattr(self.b, "mesh_opacity", not bool(animate))

                if animate:
                    fill_anim = Animation(d=0.4, mesh_opacity=1)
                    fill_anim.bind(on_progress=self.fill_up_shapes)
                    anim += fill_anim

            if animate:
                anim.cancel_all(self.b)
                anim.bind(on_progress=self.update_canvas)
                anim.start(self.b)
            elif not self.fill:
                self.update_canvas()
            else:
                self.b.canvas.clear()
                self.fill_up_shapes()

    def _calc_paths(self, animate):

        line_count = 0
        bezier_count = 0
        anim_list = []
        for id_, closed_paths in self.closed_shapes.items():
            # TODO: Support for other svg shapes like Arc, Circle, Text etc.

            for s in closed_paths[id_ + "paths"]:
                tmp = []

                for e in s:
                    if isinstance(e, Line):

                        nm_mesh = f"line{line_count}"

                        lp = line_points(
                            e, *self.b.size, *self.b.pos,
                            *self.sw_size, self.sf
                        )

                        setattr(self.b, f"{nm_mesh}_start_x", lp[0])
                        setattr(self.b, f"{nm_mesh}_start_y", lp[1])
                        setattr(self.b, f"{nm_mesh}_end_x",
                                lp[0] if animate else lp[2])
                        setattr(self.b, f"{nm_mesh}_end_y",
                                lp[1] if animate else lp[3])
                        setattr(self.b, f"{nm_mesh}_width",
                                1 if animate else self.outline_width)

                        if animate:
                            anim_list.append(
                                Animation(
                                    d=self.DUR,
                                    **{
                                        f"{nm_mesh}_end_x": lp[2],
                                        f"{nm_mesh}_end_y": lp[3],
                                        f"{nm_mesh}_width": self.outline_width
                                    }
                                )
                            )
                        line_count += 1
                        tmp.extend(lp)

                    if isinstance(e, CubicBezier):

                        nm_mesh = f"bezier{bezier_count}"

                        bp = bezier_points(
                            e, *self.b.size, *self.b.pos,
                            *self.sw_size, self.sf
                        )

                        setattr(self.b, f"{nm_mesh}_start_x", bp[0])
                        setattr(self.b, f"{nm_mesh}_start_y", bp[1])
                        setattr(self.b, f"{nm_mesh}_control1_x",
                                bp[0] if animate else bp[2])
                        setattr(self.b, f"{nm_mesh}_control1_y",
                                bp[1] if animate else bp[3])
                        setattr(self.b, f"{nm_mesh}_control2_x",
                                bp[0] if animate else bp[4])
                        setattr(self.b, f"{nm_mesh}_control2_y",
                                bp[1] if animate else bp[5])
                        setattr(self.b, f"{nm_mesh}_end_x",
                                bp[0] if animate else bp[6])
                        setattr(self.b, f"{nm_mesh}_end_y",
                                bp[1] if animate else bp[7])
                        setattr(self.b, f"{nm_mesh}_width",
                                1 if animate else self.outline_width)

                        if animate:
                            anim_list.append(
                                Animation(
                                    d=self.DUR,
                                    **{
                                        f"{nm_mesh}_control1_x": bp[2],
                                        f"{nm_mesh}_control1_y": bp[3],
                                        f"{nm_mesh}_control2_x": bp[4],
                                        f"{nm_mesh}_control2_y": bp[5],
                                        f"{nm_mesh}_end_x": bp[6],
                                        f"{nm_mesh}_end_y": bp[7],
                                        f"{nm_mesh}_width": self.outline_width
                                    }
                                )
                            )
                        bezier_count += 1
                        tmp.extend(
                            get_all_points(
                                (bp[0], bp[1]),
                                (bp[2], bp[3]),
                                (bp[4], bp[5]),
                                (bp[6], bp[7]),
                            )
                        )

                if tmp not in closed_paths[f"{id_}shapes"]:
                    closed_paths[f"{id_}shapes"].append(tmp)

        return anim_list

    def update_canvas(self, *args, **kwargs):
        self.b.canvas.clear()

        with self.b.canvas:
            Color(*self.LINE_COLOR)

            line_count = 0
            bezier_count = 0

            # Draw svg
            for e in self.path:

                if isinstance(e, Line):

                    nm_mesh = f"line{line_count}"

                    KivyLine(
                        points=[
                            getattr(self.b, f"{nm_mesh}_start_x"),
                            getattr(self.b, f"{nm_mesh}_start_y"),
                            getattr(self.b, f"{nm_mesh}_end_x"),
                            getattr(self.b, f"{nm_mesh}_end_y"),
                        ],
                        width=getattr(self.b, f"{nm_mesh}_width")
                    )

                    line_count += 1

                if isinstance(e, CubicBezier):

                    nm_mesh = f"bezier{bezier_count}"

                    KivyLine(
                        bezier=[
                            getattr(self.b, f"{nm_mesh}_start_x"),
                            getattr(self.b, f"{nm_mesh}_start_y"),
                            getattr(self.b, f"{nm_mesh}_control1_x"),
                            getattr(self.b, f"{nm_mesh}_control1_y"),
                            getattr(self.b, f"{nm_mesh}_control2_x"),
                            getattr(self.b, f"{nm_mesh}_control2_y"),
                            getattr(self.b, f"{nm_mesh}_end_x"),
                            getattr(self.b, f"{nm_mesh}_end_y"),
                        ],
                        width=getattr(self.b, f"{nm_mesh}_width"),
                    )

                    bezier_count += 1



import math
from kivy.lang.builder import Builder
from kivy.uix.widget import Widget
from kivy.properties import (
	BooleanProperty, NumericProperty,
	ListProperty, ReferenceListProperty,
)

OUTLINE_ZERO = 0.00000000001
# replaces user's 0 value for outlines, avoids invalid width exception


Builder.load_string("""

<JoystickPad>:
	id: pad
	canvas:
		Color:
			rgba: self._background_color
		Ellipse:
			pos: (self.center_x - self._radius), (self.center_y - self._radius)
			size: (self._diameter, self._diameter)

		Color:
			rgba: self._line_color
		Line:
			circle: (self.center_x, self.center_y, (self._diameter - (self._line_width * 2)) / 2)
			width: self._line_width

			
<JoyStick>:
	canvas:
		Color:
			rgba: self.outer_background_color
		Ellipse:
			pos: (self.center_x - self._outer_radius), (self.center_y - self._outer_radius)
			size: (self._outer_diameter, self._outer_diameter)

		Color:
			rgba: self.outer_line_color
		Line:
			circle: (self.center_x, self.center_y, (self._outer_radius - (self._outer_line_width / 2)))
			width: self._outer_line_width

		Color:
			rgba: self.inner_background_color
		Ellipse:
			pos: (self.center_x - self._inner_radius), (self.center_y - self._inner_radius)
			size: (self._inner_diameter, self._inner_diameter)

		Color:
			rgba: self.inner_line_color
		Line:
			circle: (self.center_x, self.center_y, self._inner_radius)
			width: self._inner_line_width

	JoystickPad:
		id: pad


""")


class TouchData:
	x_distance = None
	y_distance = None
	x_offset = None
	y_offset = None
	relative_distance = None
	is_external = None
	in_range = None

	def __init__(self, joystick, touch):
		self.joystick = joystick
		self.touch = touch
		self._calculate()

	def _calculate(self):
		js = self.joystick
		touch = self.touch
		x_distance = js.center_x - touch.x
		y_distance = js.center_y - touch.y
		x_offset = touch.x - js.center_x
		y_offset = touch.y - js.center_y
		relative_distance = ((x_distance ** 2) + (y_distance ** 2)) ** 0.5
		is_external = relative_distance > js._total_radius
		in_range = relative_distance <= js._radius_difference
		self._update(x_distance, y_distance, x_offset, y_offset,
					 relative_distance, is_external, in_range)

	def _update(self, x_distance, y_distance, x_offset, y_offset,
				relative_distance, is_external, in_range):
		self.x_distance = x_distance
		self.y_distance = y_distance
		self.x_offset = x_offset
		self.y_offset = y_offset
		self.relative_distance = relative_distance
		self.is_external = is_external
		self.in_range = in_range


class JoystickPad(Widget):
	_diameter = NumericProperty(1)
	_radius = NumericProperty(1)
	_background_color = ListProperty([0, 0, 0, 1])
	_line_color = ListProperty([1, 1, 1, 1])
	_line_width = NumericProperty(1)


class Joystick(Widget):
	'''The joystick base is comprised of an outer circle & an inner circle.
	   The joystick pad is another circle,
		   which the user can move within the base.
	   All 3 of these elements can be styled independently
		   to create different effects.
	   All coordinate properties are based on the
		   position of the joystick pad.'''


	outer_size = NumericProperty(1)
	inner_size = NumericProperty(0.75)
	pad_size = NumericProperty(0.5)
	'''Sizes are defined by percentage,
		   1.0 being 100%, of the total widget size.
		The smallest value of widget.width & widget.height
		   is used as a baseline for these percentages.'''

	outer_background_color = ListProperty([0.75, 0.75, 0.75, 1])
	inner_background_color = ListProperty([0.75, 0.75, 0.75, 1])
	pad_background_color = ListProperty([0.4, 0.4, 0.4, 1])
	'''Background colors for the joystick base & pad'''

	outer_line_color = ListProperty([0.25, 0.25, 0.25, 1])
	inner_line_color = ListProperty([0.7, 0.7, 0.7, 1])
	pad_line_color = ListProperty([0.35, 0.35, 0.35, 1])
	'''Border colors for the joystick base & pad'''

	outer_line_width = NumericProperty(0.01)
	inner_line_width = NumericProperty(0.01)
	pad_line_width = NumericProperty(0.01)
	'''Outline widths for the joystick base & pad.
	   Outline widths are defined by percentage,
		   1.0 being 100%, of the total widget size.'''

	sticky = BooleanProperty(False)
	'''When False, the joystick will snap back to center on_touch_up.
	   When True, the joystick will maintain its final position
		   at the time of on_touch_up.'''


	pad_x = NumericProperty(0.0)
	pad_y = NumericProperty(0.0)
	pad = ReferenceListProperty(pad_x, pad_y)
	'''pad values are touch coordinates in relation to
		   the center of the joystick.
	   pad_x & pad_y return values between -1.0 & 1.0.
	   pad returns a tuple of pad_x & pad_y, and is the best property to
		   bind to in order to receive updates from the joystick.'''

	@property
	def magnitude(self):
		return self._magnitude
	'''distance of the pad, between 0.0 & 1.0,
		   from the center of the joystick.'''

	@property
	def radians(self):
		return self._radians
	'''degrees of the pad, between 0.0 & 360.0, in relation to the x-axis.'''

	@property
	def angle(self):
		return math.degrees(self.radians)
	'''position of the pad in radians, between 0.0 & 6.283,
		   in relation to the x-axis.'''

	'''magnitude, radians, & angle can be used to
		   calculate polar coordinates'''


	_outer_line_width = NumericProperty(OUTLINE_ZERO)
	_inner_line_width = NumericProperty(OUTLINE_ZERO)
	_pad_line_width = NumericProperty(OUTLINE_ZERO)

	_total_diameter = NumericProperty(0)
	_total_radius = NumericProperty(0)

	_inner_diameter = NumericProperty(0)
	_inner_radius = NumericProperty(0)

	_outer_diameter = NumericProperty(0)
	_outer_radius = NumericProperty(0)

	_magnitude = 0

	@property
	def _radians(self):
		if not (self.pad_y and self.pad_x):
			return 0
		arc_tangent = math.atan(self.pad_y / self.pad_x)
		if self.pad_x > 0 and self.pad_y > 0:    # 1st Quadrant
			return arc_tangent
		elif self.pad_x > 0 and self.pad_y < 0:  # 4th Quadrant
			return (math.pi * 2) + arc_tangent
		else:                                    # 2nd & 3rd Quadrants
			return math.pi + arc_tangent

	@property
	def _radius_difference(self):
		return (self._total_radius - self.ids.pad._radius)

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		self.bind(
			pos=self.do_layout,
			size=self.do_layout,
			children=self.do_layout
		)

	def move_pad(self, touch, from_touch_down):
		td = TouchData(self, touch)
		if td.is_external and from_touch_down:
			touch.ud['joystick'] = None
			return False
		elif td.in_range:
			self._update_coordinates_from_internal_touch(touch, td)
			return True
		elif not (td.in_range):
			self._update_coordinates_from_external_touch(td)
			return True

	def center_pad(self):
		self.ids.pad.center = self.center
		self._magnitude = 0
		self.pad_x = 0
		self.pad_y = 0

	def _update_coordinates_from_external_touch(self, touchdata):
		td = touchdata
		pad_distance = self._radius_difference * (1.0 / td.relative_distance)
		x_distance_offset = -td.x_distance * pad_distance
		y_distance_offset = -td.y_distance * pad_distance
		x = self.center_x + x_distance_offset
		y = self.center_y + y_distance_offset
		radius_offset = pad_distance / self._radius_difference
		self.pad_x = td.x_offset * radius_offset
		self.pad_y = td.y_offset * radius_offset
		self._magnitude = 1.0
		self.ids.pad.center = (x, y)

	def _update_coordinates_from_internal_touch(self, touch, touchdata):
		td = touchdata
		self.pad_x = td.x_offset / self._radius_difference
		self.pad_y = td.y_offset / self._radius_difference
		self._magnitude = td.relative_distance / \
			(self._total_radius - self.ids.pad._radius)
		self.ids.pad.center = (touch.x, touch.y)
	

	def do_layout(self, *args):
		if 'pad' in self.ids:
			size = min(*self.size)
			self._update_outlines(size)
			self._update_circles(size)
			self._update_pad()

	def _update_outlines(self, size):
		self._outer_line_width = (self.outer_line_width * size) \
			if (self.outer_line_width) else (OUTLINE_ZERO)
		self._inner_line_width = (self.inner_line_width * size) \
			if (self.inner_line_width) else (OUTLINE_ZERO)
		self.ids.pad._line_width = (self.pad_line_width * size) \
			if (self.pad_line_width) else (OUTLINE_ZERO)

	def _update_circles(self, size):
		self._total_diameter = size
		self._total_radius = self._total_diameter / 2
		self._outer_diameter = \
			(self._total_diameter - self._outer_line_width) * self.outer_size
		self._outer_radius = self._outer_diameter / 2
		self.ids.pad._diameter = self._total_diameter * self.pad_size
		self.ids.pad._radius = self.ids.pad._diameter / 2
		self._inner_diameter = \
			(self._total_diameter - self._inner_line_width) * self.inner_size
		self._inner_radius = self._inner_diameter / 2

	def _update_pad(self):
		self.ids.pad.center = self.center
		self.ids.pad._background_color = self.pad_background_color
		self.ids.pad._line_color = self.pad_line_color


	def on_touch_down(self, touch):
		if self.collide_point(touch.x, touch.y):
			touch.ud['joystick'] = self
			return self.move_pad(touch, from_touch_down=True)
		return super().on_touch_down(touch)

	def on_touch_move(self, touch):
		if self._touch_is_active(touch):
			return self.move_pad(touch, from_touch_down=False)
		return super().on_touch_move(touch)

	def on_touch_up(self, touch):
		if self._touch_is_active(touch) and not (self.sticky):
			self.center_pad()
			return True
		return super().on_touch_up(touch)

	def _touch_is_active(self, touch):
		return 'joystick' in touch.ud and touch.ud['joystick'] == self




from kivy.properties import (
	ObjectProperty, NumericProperty,
	StringProperty, ListProperty,
	BooleanProperty, ColorProperty
)

from kivy.graphics import (
	Fbo, ClearBuffers, ClearColor,
	Scale, Translate
)

from kivy.uix.widget import Widget
from kivy.lang.builder import Builder
from kivy.core.window import Window
from kivy.metrics import dp
from kivy.clock import Clock


Builder.load_string("""

<SpacerWidget>:
	size_hint: [None, None]
	size: [0, 0]
	canvas:
		Color:
			rgba: self.outline_color
		Line:
			rectangle: [*self.pos, *self.size]
			width: root.outline_width

<PreviewWidget>:
	size_hint: [None, None]
	size: [0, 0]
	canvas:
		Color:
			rgba: self.outline_color
		Line:
			rectangle: [\
				self.to_window(self.x, 0)[0] - root.outline_width, \
				self.to_window(0, self.y)[1] - root.outline_width, \
				self.width + (root.outline_width * 2), \
				self.height + (root.outline_width * 2)\
				]
			width: root.outline_width
		Color:
			rgba: self.background_color
		Rectangle:
			pos: self.pos
			size: self.size
			texture: self.texture

""")


class SpacerWidget(Widget):
	outline_width = NumericProperty(dp(2))
	outline_color = ColorProperty([1, 0, 0, 1])


class PreviewWidget(Widget):
	texture = ObjectProperty(None)
	outline_width = NumericProperty(dp(1.01))
	outline_color = ColorProperty([1, 0, 0, 1])
	background_color = ColorProperty([1, 1, 1, 1])


class DraggableObjectBehavior(Widget):

	drag_cls = StringProperty('')

	_drag_touch = None

	drag_distance = NumericProperty(dp(10))

	preview_widget = ObjectProperty(None)

	touch_dx = NumericProperty(0)
	touch_dy = NumericProperty(0)

	dragging = BooleanProperty(False)
 
	start_widget_pos = ListProperty([0, 0])


	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		Clock.schedule_once(self.start)
	
	def start(self, *args):
		if self.preview_widget == None:
			self.preview_widget = PreviewWidget()

	def _touch_uid(self):
		return '{}.{}'.format(self.__class__.__name__, self.uid)

	def on_touch_down(self, touch):
		uid = self._touch_uid()
		if uid in touch.ud:
			return touch.ud[uid]

		if super().on_touch_down(touch):
			touch.ud[uid] = False
			return True

		x, y = touch.pos
		if not self.collide_point(x, y):
			touch.ud[uid] = False
			return False

		if self._drag_touch or ('button' in touch.profile and
								touch.button.startswith('scroll')):
			touch.ud[uid] = False
			return False

		self._drag_touch = touch
		touch.grab(self)
		touch.ud[uid] = True

		return self.drag_down(touch)

	def on_touch_move(self, touch):
		uid = self._touch_uid()
		if uid not in touch.ud:
			touch.ud[uid] = False
			return super().on_touch_move(touch)

		if not touch.ud[uid]:
			return super().on_touch_move(touch)

		if touch.grab_current is not self:
			return False

		return self.drag_move(touch)

	def on_touch_up(self, touch):
		uid = self._touch_uid()
		if uid not in touch.ud:
			touch.ud[uid] = False
			return super().on_touch_up(touch)

		if not touch.ud[uid]:
			return super().on_touch_up(touch)

		if touch.grab_current is not self:
			return False

		touch.ungrab(self)
		self._drag_touch = None

		return self.drag_up(touch)

	def drag_down(self, touch):
		self.clean_dragging()
		self.prepare_preview_widget(touch)
		self.touch_dx = self.touch_dy = 0
		self.dragging = False
		self.preview_widget.canvas.opacity = 0
		
		Window.add_widget(self.preview_widget)
		self.start_widget_pos = self.to_window(*self.pos)
		self.preview_widget.pos = self.start_widget_pos
		return False

	def drag_move(self, touch):
		if not self.dragging:
			self.touch_dx += abs(touch.dx)
			self.touch_dy += abs(touch.dy)

			if (self.touch_dx ** 2 + self.touch_dy ** 2) ** 0.5 > self.drag_distance:
				self.dragging = True
				self.preview_widget.canvas.opacity = 0.4
				touch.ud['drag_cls'] = self.drag_cls
				touch.ud['drag_widget'] = self
				self.parent.remove_widget(self)
			else:
				return False

		x, y = self.start_widget_pos
		x += (touch.x - touch.ox)
		y += (touch.y - touch.oy)
		self.preview_widget.pos = [x, y]
		return False

	def drag_up(self, touch):
		self.clean_dragging()
		if not self.dragging:
			return False

		self.clean_dragging()
		self.dragging = False
		return False

	def prepare_preview_widget(self, touch):
		fbo = Fbo(size=self.size, with_stencilbuffer=True)
		with fbo:
			ClearColor(0, 0, 0, 0)
			ClearBuffers()
			Scale(1, -1, 1)
			Translate(-self.x, -self.y - self.height, 0)

		fbo.add(self.canvas)
		fbo.draw()
		fbo.texture.flip_vertical()

		self.preview_widget.size = self.size
		self.preview_widget.texture = fbo.texture

	def clean_dragging(self):
		"""Removes the drag widget preview.
		"""
		self.preview_pixels = None
		if self.preview_widget.parent:
			self.preview_widget.parent.remove_widget(self.preview_widget)


class DraggableLayoutBehavior(Widget):

	spacer_widget = ObjectProperty(None)

	drag_classes = ListProperty([])

	drag_append_end = BooleanProperty(False)

	def __init__(self, *args, **kwargs):
		self.spacer_widget = SpacerWidget()
		super().__init__(*args, **kwargs)

	def _touch_uid(self):
		return '{}.{}'.format(self.__class__.__name__, self.uid)

	def compare_pos_to_widget(self, widget, pos):
		return 'before'

	def get_widget_under_drag(self, x, y):
		for widget in self.children:
			if widget.collide_point(x, y):
				return widget
		
		return None

	def handle_drag_release(self, index, drag_widget, touch):
		self.add_widget(drag_widget, index)

	def get_drop_insertion_index_move(self, x, y):
		spacer = self.spacer_widget
		widget = self.get_widget_under_drag(x, y)

		if widget in (spacer, None):
			return None

		i = self.children.index(widget)
		j = None
		if self.compare_pos_to_widget(widget, (x, y)) == 'before':
			if i == len(self.children) - 1 or self.children[i + 1] != spacer:
				j = i + 1
		else:
			if not i or self.children[i - 1] != spacer:
				j = i
		return j

	def get_drop_insertion_index_up(self, x, y):
		spacer = self.spacer_widget
		widget = self.get_widget_under_drag(x, y)
		
		if widget == spacer:
			index = self.children.index(spacer)
		elif widget == None:
			index = 0
		else:
			if self.compare_pos_to_widget(widget, (x, y)) == 'before':
				index = self.children.index(widget) + 1
			else:
				index = self.children.index(widget)

		if spacer.parent:
			i = self.children.index(spacer)
			self.remove_widget(spacer)
			if i < index:
				index -= 1
		
		return index

	def on_touch_move(self, touch):
		spacer = self.spacer_widget
		preview_widget = touch.ud.get('drag_widget')
		if preview_widget != None:
			spacer.size = preview_widget.size

			pw = preview_widget.preview_widget
			spacer.outline_color = pw.outline_color
			spacer.outline_width = pw.outline_width
		
		x, y = touch.pos
		if touch.grab_current is not self:
			if not touch.ud.get(self._touch_uid()):
				# we haven't dealt with this before
				if self.collide_point(*touch.pos) and \
						touch.ud.get('drag_cls') in self.drag_classes:
					if super().on_touch_move(touch):
						return True
					touch.grab(self)
					touch.ud[self._touch_uid()] = True
				else:
					return super().on_touch_move(touch)
			else:
				# we have dealt with this touch before, do it when grab_current
				return True
		else:
			if touch.ud.get('drag_cls') not in self.drag_classes or \
					not self.collide_parent_tree(x, y):
				touch.ungrab(self)
				del touch.ud[self._touch_uid()]
				if spacer.parent:
					self.remove_widget(spacer)
				return False
			if super().on_touch_move(touch):
				touch.ungrab(self)
				del touch.ud[self._touch_uid()]
				if spacer.parent:
					self.remove_widget(spacer)
				return True

		if self.drag_append_end:
			if not spacer.parent:
				self.add_widget(spacer)
			return True

		j = self.get_drop_insertion_index_move(x, y)
		if j != None:
			if spacer.parent:
				i = self.children.index(spacer)
				self.remove_widget(spacer)
				if i < j:
					j -= 1
			self.add_widget(spacer, index=j)
		
		return True

	def on_touch_up(self, touch):
		spacer = self.spacer_widget
		if touch.grab_current is not self:
			if not touch.ud.get(self._touch_uid()):
				# we haven't dealt with this before
				if self.collide_point(*touch.pos) and \
						touch.ud.get('drag_cls') in self.drag_classes:
					if super().on_touch_up(touch):
						return True
					x, y = touch.pos
				else:
					return super().on_touch_up(touch)
			else:
				# we have dealt with this touch before, do it when grab_current
				return True
		else:
			touch.ungrab(self)
			del touch.ud[self._touch_uid()]
			x, y = touch.pos
			if touch.ud.get('drag_cls') not in self.drag_classes or \
					not self.collide_parent_tree(x, y):
				if spacer.parent:
					self.remove_widget(spacer)
				return False
			if super().on_touch_up(touch):
				if spacer.parent:
					self.remove_widget(spacer)
				return True

		if self.drag_append_end:
			if spacer.parent:
				self.remove_widget(spacer)
			
			self.handle_drag_release(0, touch.ud['drag_widget'], touch)
			return True

		index = self.get_drop_insertion_index_up(x, y)
		self.handle_drag_release(index, touch.ud['drag_widget'], touch)
		return True
	
	def collide_parent_tree(self, x, y):
		"""Returns whether (x, y) collide with the widget and all its parents.
		"""
		if not self.collide_point(x, y):
			return False

		parent = self.parent
		while parent and hasattr(parent, 'to_parent'):
			x, y = parent.to_parent(x, y)  # transform to parent's parent's
			if not parent.collide_point(x, y):
				return False

			parent = parent.parent
		
		return True





from kivy.uix.anchorlayout import AnchorLayout
from kivy.properties import ListProperty, OptionProperty, NumericProperty
from kivy.clock import Clock
from kivy.lang import Builder
from kivygo.behaviors.touch_effecs import EffectBehavior
from kivygo.behaviors.hover import HoverBehavior


Builder.load_string("""


<ColoredAnchorLayout>:
	canvas.before: 
		Color:
			rgba: self._background_color
		RoundedRectangle:
			pos: self.pos
			size: self.size
			radius: self.radius
	canvas.after:
		Color:
			rgba: self.stroke_color
		Line:
			rounded_rectangle: [*self.pos, *self.size, *self.radius]
			width: self.stroke_width

<AnchorLayoutBaseBack>:
	anchor_x: 'center'
	anchor_y: 'center'
	canvas.before:
		Color:
			rgba: self._background_color
		RoundedRectangle:
			size: self.size
			pos: self.pos
			radius: self.radius

""")


class ColoredAnchorLayout(AnchorLayout, HoverBehavior, EffectBehavior):

	stroke_color = ListProperty([0, 0 ,0 ,0])
	stroke_width = NumericProperty(2)

	background_color = ListProperty([0, 0, 0, 0])
	background_color_pos = ListProperty([0, 0, 0, 0])
	
	_background_color = ListProperty([0, 0, 0, 0])

	radius = ListProperty([0, 0, 0, 0])
	effect_color = ListProperty([0, 0, 0, 0])

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self.type_button = 'rounded'
		Clock.schedule_once(self.set_color)
		self.bind(background_color=self.set_color)
	
	def set_color(self, *args):
		if isinstance(self.background_color[0], (int, float)):

			if self.background_color != [0, 0, 0, 0]:
				self._background_color = self.background_color

		elif self.background_color[0] != [0, 0, 0, 0]:
			self._background_color = self.background_color[0]
		
		if isinstance(self.background_color_pos[0], (int, float)):

			if self.background_color_pos != [0, 0, 0, 0]:
				self._background_color = self.background_color_pos
				
		elif self.background_color_pos[0] != [0, 0, 0, 0]:
			self._background_color = self.background_color_pos[0]

	def on_cursor_enter(self, *args):
		if not isinstance(self.background_color_pos[0], (int, float)):
			self._background_color = self.background_color_pos[1]

		return super().on_cursor_enter(*args)

	def on_cursor_leave(self, *args):
		if not isinstance(self.background_color_pos[0], (int, float)):
			self._background_color = self.background_color_pos[0]
			
		return super().on_cursor_leave(*args)

	def on_touch_down(self, touch):
		if not self.collide_point(*touch.pos):
			return False
		
		if len(self.background_color) == 2 and isinstance(self.background_color, (list, tuple)):
			self._background_color = self.background_color[1]
		
		self.ripple_show(touch)        
		return super().on_touch_down(touch)

	def on_touch_up(self, touch):
		if not self.collide_point(*touch.pos):
			return False
		
		if len(self.background_color) == 2 and isinstance(self.background_color, (list, tuple)):
			self._background_color = self.background_color[0]

		self.ripple_fade()
		return super().on_touch_up(touch)



class AnchorLayoutButton(ColoredAnchorLayout):
	
	state = OptionProperty('normal', options=('normal', 'down'))
	
	def set_color(self, *args):
		super().set_color(*args)
		Clock.schedule_once(lambda *a: self.on_state(self, self.state))


	def on_state(self, widget, state):
		toggle = None if not self.children else self.children[0]
		if state == "normal" and hasattr(toggle, "group"):
			if (not toggle.allow_no_selection and toggle.group and toggle.state == 'down'):
				self.state = 'down'
				return None

		if state == "down":
			if not isinstance(self.background_color[0], (int, float)):
				self._background_color = self.background_color[1]
			return None
			
		if not isinstance(self.background_color[0], (int, float)):
			self._background_color = self.background_color[0]


	def on_touch_down(self, touch):
		vl = super().on_touch_down(touch)
		if not vl: return False
		
		toggle = None if not self.children else self.children[0]
		if hasattr(toggle, "group"):
			self.state = 'normal' if self.state == 'down' else 'down'
		else:
			self.state = 'down'

		if hasattr(toggle, "state") and not vl:
			if hasattr(toggle, "group"):
				toggle._do_press()
			toggle.state = self.state
		
		return vl

	def on_touch_up(self, touch):
		vl = super().on_touch_up(touch)
		
		if not vl and hasattr(self.children[0], "group"):
			self.state = self.children[0].state
		elif vl:
			self.state = "normal"
		
		if not vl:
			return False
		
		if not vl and hasattr(self.children[0], "group"):
			self.children[0]._do_release()
			self.state = self.children[0].state
		elif not vl and hasattr(self.children[0], "state"):
			self.children[0].state = self.state

		return vl

